# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# An alternate Python Minecraft library for the Rasperry-Pi
# Copyright (c) 2013-2015 Dave Jones <dave@waveform.org.uk>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
str = type('')


import pytest
import socket
import select
try:
    from unittest import mock
except ImportError:
    import mock
from picraft import Connection, ConnectionError, CommandError, BatchStarted, BatchNotStarted


def test_connection_init_pi():
    with mock.patch('socket.socket'), mock.patch('select.select'):
        select.select.return_value = [False]
        conn = Connection('myhost', 1234)
        conn._socket.connect.assert_called_once_with(('myhost', 1234))
        assert conn.server_version == 'minecraft-pi'

def test_connection_init_juice():
    with mock.patch('socket.socket'), mock.patch('select.select'):
        select.select.return_value = [True]
        mock_sock = socket.socket()
        mock_file = mock_sock.makefile()
        mock_file.readline.return_value = b'Fail\n'
        conn = Connection('myhost', 1234)
        conn._socket.connect.assert_called_once_with(('myhost', 1234))
        assert conn.server_version == 'raspberry-juice'

def test_connection_init_unknown():
    with mock.patch('socket.socket'), mock.patch('select.select'):
        select.select.return_value = [True]
        mock_sock = socket.socket()
        mock_file = mock_sock.makefile()
        mock_file.readline.return_value = b'bar\n'
        with pytest.raises(CommandError):
            conn = Connection('myhost', 1234)

def test_connection_close():
    with mock.patch('socket.socket'), mock.patch('select.select'):
        select.select.return_value = [False]
        conn = Connection('myhost', 1234)
        s = conn._socket
        conn.close()
        # Repeated attempts shouldn't fail
        conn.close()
        s.close.assert_called_once_with()

def test_connection_send():
    with mock.patch('socket.socket'), mock.patch('select.select'):
        select.select.return_value = [False]
        conn = Connection('myhost', 1234)
        conn._wfile.write.reset_mock()
        conn.send('foo()')
        conn._wfile.write.assert_called_once_with(b'foo()\n')

def test_connection_send_error():
    with mock.patch('socket.socket'), mock.patch('select.select'):
        select.select.side_effect = [[False], [True]]
        conn = Connection('myhost', 1234)
        conn._rfile.readline.return_value = b'Fail\n'
        with pytest.raises(ConnectionError):
            conn.send('foo()')

def test_connection_transact():
    with mock.patch('socket.socket'), mock.patch('select.select'):
        select.select.side_effect = [[False], [True]]
        conn = Connection('myhost', 1234)
        conn._wfile.write.reset_mock()
        conn._rfile.readline.return_value = b'bar\n'
        result = conn.transact('foo()')
        conn._wfile.write.assert_called_once_with(b'foo()\n')
        assert result == 'bar'

def test_connection_batch_send():
    with mock.patch('socket.socket'), mock.patch('select.select'):
        select.select.return_value = [False]
        conn = Connection('myhost', 1234)
        conn._wfile.write.reset_mock()
        with conn.batch_start():
            conn.send('foo()')
            conn.send('bar()')
            conn.send('baz()')
        conn._wfile.write.assert_called_once_with(b'foo()\nbar()\nbaz()\n')

def test_connection_batch_forget():
    with mock.patch('socket.socket'), mock.patch('select.select'):
        select.select.return_value = [False]
        conn = Connection('myhost', 1234)
        conn._wfile.write.reset_mock()
        conn.batch_start()
        conn.send('foo()')
        conn.send('bar()')
        conn.send('baz()')
        conn.batch_forget()
        assert not conn._wfile.write.called

def test_connection_batch_exception():
    with mock.patch('socket.socket'), mock.patch('select.select'):
        select.select.return_value = [False]
        conn = Connection('myhost', 1234)
        conn._wfile.write.reset_mock()
        try:
            with conn.batch_start():
                conn.send('foo()')
                conn.send('bar()')
                conn.send('baz()')
                raise Exception('boo')
        except Exception:
            pass
        assert not conn._wfile.write.called

def test_connection_batch_start_fail():
    with mock.patch('socket.socket'), mock.patch('select.select'):
        select.select.return_value = [False]
        conn = Connection('myhost', 1234)
        with pytest.raises(BatchStarted):
            conn.batch_start()
            conn.batch_start()

def test_connection_batch_send_fail():
    with mock.patch('socket.socket'), mock.patch('select.select'):
        select.select.return_value = [False]
        conn = Connection('myhost', 1234)
        with pytest.raises(BatchNotStarted):
            conn.batch_send()

def test_connection_ignore_errors():
    with mock.patch('socket.socket'), mock.patch('select.select'):
        select.select.side_effect = [[False], [True], [False]]
        conn = Connection('myhost', 1234, ignore_errors=True)
        conn.send('foo()')
        conn._socket.recv.assert_called_once_with(1500)




# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# An alternate Python Minecraft library for the Rasperry-Pi
# Copyright (c) 2013-2015 Dave Jones <dave@waveform.org.uk>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
The picraft package consists of several modules which permit access to and
modification of a Minecraft world. The package is intended as an alternative
Python API to the "official" Minecraft Python API (for reasons explained in the
:ref:`faq`).

The classes defined in most modules of this package are available directly
from the :mod:`picraft` namespace. In other words, the following code is
typically all that is required to access classes in this package::

    import picraft

For convenience on the command line you may prefer to simply do the following::

    from picraft import *

However, this is frowned upon in code as it pulls everything into the global
namespace, so you may prefer to do something like this::

    from picraft import World, Vector, Block

This is the style used in the :ref:`recipes` chapter. Sometimes, if you are
using the :class:`~picraft.vector.Vector` class extensively, you may wish to
use the short-cuts for it::

    from picraft import World, V, O, X, Y, Z, Block

The following sections document the various modules available within the
package:

* :ref:`api_world`
* :ref:`api_block`
* :ref:`api_vector`
* :ref:`api_events`
* :ref:`api_connection`
* :ref:`api_player`
* :ref:`api_exc`
"""

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )

# Make Py2's str equivalent to Py3's
str = type('')


from .exc import (
    Error,
    ConnectionError,
    CommandError,
    NoResponse,
    BatchStarted,
    BatchNotStarted,
    NotSupported,
    ConnectionClosed,
    EmptySliceWarning,
    )
from .vector import Vector, vector_range, line, lines, circle, V, O, X, Y, Z
from .block import Block, BLOCK_COLORS
from .events import BlockHitEvent
from .connection import Connection
from .player import Players, Player, HostPlayer
from .world import World




# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# An alternate Python Minecraft library for the Rasperry-Pi
# Copyright (c) 2013-2015 Dave Jones <dave@waveform.org.uk>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
str = type('')


MAX_VALUE = (2 - 2 ** -52) * 2 ** 1023

# See <http://floating-point-gui.de/errors/comparison/> for the base
# functionality below. This function is a variant of that presented in the site
# intended to ignore the (quite large) errors in the trigonometry functions
# near zero.
#
# For example math.cos(math.pi/2) should be 0, but outputs something close to
# 6e-17. This cannot be compared to 0 with the function presented on the
# aforementioned site unless epsilon is set to an insanely high value (close to
# 1e300). Instead, in this function we subtract x and y from one and compare
# this result instead. This is a valid practice given that the range of the
# functions is 0 to 1, and given that we never expect to compare extremely two
# deliberately extremely small values with this function.
#
# WARNING: if this assumption is violated, tests may incorrectly pass when they
# should fail.

def fp_equal(x, y, epsilon=1e-9):
    if x == y:
        return True
    if x == 0.0 or y == 0.0:
        x = 1.0 - x
        y = 1.0 - y
    # use relative error
    return abs(x - y) / min(abs(x) + abs(y), MAX_VALUE) < epsilon

def fp_vectors_equal(vx, vy, epsilon=1e-9):
    return (
            fp_equal(vx.x, vy.x, epsilon) and
            fp_equal(vx.y, vy.y, epsilon) and
            fp_equal(vx.z, vy.z, epsilon))




# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# An alternate Python Minecraft library for the Rasperry-Pi
# Copyright (c) 2013-2015 Dave Jones <dave@waveform.org.uk>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
str = type('')


import pytest
import io
import picraft.world
import picraft.block
import picraft.player
import picraft.events
from picraft import World, Vector, vector_range, Connection, NotSupported
try:
    from unittest import mock
except ImportError:
    import mock


def test_world_init():
    with mock.patch('picraft.world.Connection') as c:
        World()
        c.assert_called_once_with('localhost', 4711, 0.3, False)

def test_world_objects():
    with mock.patch('picraft.world.Connection') as c:
        w = World()
        assert w.connection is c()
        assert isinstance(w.players, picraft.player.Players)
        assert isinstance(w.player, picraft.player.HostPlayer)
        assert isinstance(w.blocks, picraft.block.Blocks)
        assert isinstance(w.height, picraft.world.WorldHeight)
        assert isinstance(w.camera, picraft.world.Camera)
        assert isinstance(w.checkpoint, picraft.world.Checkpoint)
        assert isinstance(w.events, picraft.events.Events)

def test_world_say():
    with mock.patch('picraft.world.Connection') as c:
        World().say('Hello world!')
        c().send.assert_called_once_with('chat.post(Hello world!)')
        c().send.reset_mock()
        World().say('Hello\nworld!')
        c().send.assert_any_call('chat.post(Hello)')
        c().send.assert_any_call('chat.post(world!)')

def test_world_context():
    with mock.patch('picraft.world.Connection') as c:
        with World():
            pass
        c().close.assert_called_once_with()

def test_world_immutable():
    with mock.patch('picraft.world.Connection') as c:
        w = World()
        with pytest.raises(AttributeError):
            w.immutable
        c().server_version = 'minecraft-pi'
        w.immutable = False
        c().send.assert_called_once_with('world.setting(world_immutable,0)')
        c().server_version = 'raspberry-juice'
        with pytest.raises(NotSupported):
            w.immutable = False

def test_world_nametags():
    with mock.patch('picraft.world.Connection') as c:
        w = World()
        with pytest.raises(AttributeError):
            w.nametags_visible
        c().server_version = 'minecraft-pi'
        w.nametags_visible = False
        c().send.assert_called_once_with('world.setting(nametags_visible,0)')
        c().server_version = 'raspberry-juice'
        with pytest.raises(NotSupported):
            w.nametags_visible = False

def test_world_height_get():
    with mock.patch('picraft.world.Connection') as c:
        World().height[Vector(1, 2, 3)]
        c().transact.assert_called_once_with('world.getHeight(1,3)')
        c().transact.reset_mock()
        v_from = Vector(1, 2, 3)
        v_to = Vector(2, 3, 5)
        World().height[v_from:v_to]
        for v in vector_range(v_from, v_to):
            c().transact.assert_any_call('world.getHeight(%d,%d)' % (v.x, v.z))

def test_checkpoint_save():
    with mock.patch('picraft.world.Connection') as c:
        c().server_version = 'minecraft-pi'
        World().checkpoint.save()
        c().send.assert_called_once_with('world.checkpoint.save()')
        c().server_version = 'raspberry-juice'
        with pytest.raises(NotSupported):
            World().checkpoint.save()

def test_checkpoint_restore():
    with mock.patch('picraft.world.Connection') as c:
        c().server_version = 'minecraft-pi'
        World().checkpoint.restore()
        c().send.assert_called_once_with('world.checkpoint.restore()')
        c().server_version = 'raspberry-juice'
        with pytest.raises(NotSupported):
            World().checkpoint.restore()

def test_checkpoint_context():
    with mock.patch('picraft.world.Connection') as c:
        c().server_version = 'minecraft-pi'
        with World().checkpoint:
            pass
        c().send.assert_called_with('world.checkpoint.save()')
        try:
            with World().checkpoint:
                c().send.assert_called_with('world.checkpoint.save()')
                raise ValueError()
        except ValueError:
            c().send.assert_called_with('world.checkpoint.restore()')

def test_camera_pos():
    with mock.patch('picraft.world.Connection') as c:
        w = World()
        with pytest.raises(AttributeError):
            w.camera.pos
        c().server_version = 'minecraft-pi'
        w.camera.pos = Vector(1, 2, 3)
        assert c().send.mock_calls == [
                mock.call('camera.mode.setFixed()'),
                mock.call('camera.setPos(1,2,3)'),
                ]
        c().server_version = 'raspberry-juice'
        with pytest.raises(NotSupported):
            w.camera.pos = Vector(1, 2, 3)

def test_camera_1st_person():
    with mock.patch('picraft.world.Connection') as c:
        w = World()
        c().transact.return_value = '1|2|3'
        c().server_version = 'minecraft-pi'
        w.camera.first_person(w.player)
        c().send.assert_called_with('camera.mode.setNormal()')
        w.camera.first_person(w.players[3])
        c().send.assert_called_with('camera.mode.setNormal(3)')
        c().server_version = 'raspberry-juice'
        with pytest.raises(NotSupported):
            w.camera.first_person(w.player)

def test_camera_3rd_person():
    with mock.patch('picraft.world.Connection') as c:
        w = World()
        c().transact.return_value = '1|2|3'
        c().server_version = 'minecraft-pi'
        w.camera.third_person(w.player)
        c().send.assert_called_with('camera.mode.setFollow()')
        w.camera.third_person(w.players[3])
        c().send.assert_called_with('camera.mode.setFollow(3)')
        c().server_version = 'raspberry-juice'
        with pytest.raises(NotSupported):
            w.camera.third_person(w.player)



# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# An alternate Python Minecraft library for the Rasperry-Pi
# Copyright (c) 2013-2015 Dave Jones <dave@waveform.org.uk>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
str = type('')


import pytest
import math
from .conftest import fp_equal, fp_vectors_equal
from picraft import Vector, vector_range, line, lines, O, X, Y, Z
from picraft.vector import rmod, rdiv, sign
from picraft.compat import range


def test_vector_init():
    assert Vector() == (0, 0, 0)
    assert Vector(1) == (1, 0, 0)
    assert Vector(1, 2, 3) == (1, 2, 3)
    assert Vector(z=3, y=2, x=1) == (1, 2, 3)

def test_vector_from_string():
    assert Vector.from_string('1,2,3') == Vector(1, 2, 3)
    assert Vector.from_string('1, 2, 3') == Vector(1, 2, 3)
    with pytest.raises(ValueError):
        Vector.from_string('1')
    with pytest.raises(ValueError):
        Vector.from_string('(1,2,3)')

def test_vector_str():
    assert str(Vector()) == '0,0,0'
    assert str(Vector(1, 2, 3)) == '1,2,3'
    assert str(Vector(0.1, 0.2, 0.3)) == '0.1,0.2,0.3'

def test_vector_add():
    assert Vector() + Vector(1, 2, 3) == Vector(1,2, 3)
    assert Vector() + 1 == Vector(1, 1, 1)
    assert 1 + Vector() == Vector(1, 1, 1)

def test_vector_sub():
    assert Vector(1, 2, 3) - Vector(1, 1, 1) == Vector(0, 1, 2)
    assert Vector(1, 2, 3) - 1 == Vector(0, 1, 2)
    with pytest.raises(TypeError):
        1 - Vector()

def test_vector_mul():
    assert Vector(1, 2, 3) * Vector(3, 3, 3) == Vector(3, 6, 9)
    assert Vector(1, 2, 3) * 3 == Vector(3, 6, 9)
    assert 3 * Vector(1, 2, 3) == Vector(3, 6, 9)

def test_vector_truediv():
    assert fp_vectors_equal(Vector(1, 2, 3) / Vector(2, 2, 2), Vector(0.5, 1, 1.5))
    assert fp_vectors_equal(Vector(1, 2, 3) / 2, Vector(0.5, 1, 1.5))
    with pytest.raises(TypeError):
        2 / Vector(1, 2, 3)

def test_vector_floordiv():
    assert Vector(1, 2, 3) // Vector(2, 2, 2) == Vector(0, 1, 1)
    assert Vector(1, 2, 3) // 2 == Vector(0, 1, 1)
    with pytest.raises(TypeError):
        2 // Vector(1, 2, 3)

def test_vector_mod():
    assert Vector(1, 3, 9) % Vector(3, 3, 3) == Vector(1, 0, 0)
    assert Vector(2, 3, 4) % 2 == Vector(0, 1, 0)
    with pytest.raises(TypeError):
        2 % Vector(1, 1, 1)

def test_vector_pow():
    assert Vector(1, 2, 3) ** Vector(2, 2, 2) == Vector(1, 4, 9)
    assert Vector(1, 2, 3) ** Vector(2, 2, 2) == Vector(1, 4, 9)
    assert pow(Vector(1, 2, 3), Vector(2, 2, 2), Vector(5, 5, 5)) == Vector(1, 4, 4)
    assert Vector(1, 2, 3) ** 2 == Vector(1, 4, 9)
    assert pow(Vector(1, 2, 3), 2, 5) == Vector(1, 4, 4)
    with pytest.raises(TypeError):
        2 ** Vector(1, 2, 3)

def test_vector_lshift():
    assert Vector(1, 2, 3) << Vector(1, 2, 3) == Vector(2, 8, 24)
    assert Vector(1, 2, 3) << 1 == Vector(2, 4, 6)
    with pytest.raises(TypeError):
        2 << Vector(1, 2, 3)

def test_vector_rshift():
    assert Vector(2, 8, 16) >> Vector(0, 1, 2) == Vector(2, 4, 4)
    assert Vector(2, 8, 16) >> 1 == Vector(1, 4, 8)
    with pytest.raises(TypeError):
        2 >> Vector(1, 1, 1)

def test_vector_and():
    assert Vector(1, 2, 3) & Vector(1, 1, 1) == Vector(1, 0, 1)
    assert Vector(4, 4, 4) & Vector(0x7fff, 0, 0) == Vector(4, 0, 0)
    assert Vector(1, 2, 3) & 1 == Vector(1, 0, 1)

def test_vector_xor():
    assert Vector(1, 2, 3) ^ Vector(1, 1, 1) == Vector(0, 3, 2)
    assert Vector(4, 4, 4) ^ Vector(4, 4, 4) == Vector(0, 0, 0)
    assert Vector(1, 2, 3) ^ 1 == Vector(0, 3, 2)

def test_vector_or():
    assert Vector(1, 2, 3) | Vector(1, 1, 1) == Vector(1, 3, 3)
    assert Vector(4, 4, 4) | Vector(1, 2, 3) == Vector(5, 6, 7)
    assert Vector(1, 2, 3) | 1 == Vector(1, 3, 3)

def test_vector_neg():
    assert -Vector(1, 1, 1) == Vector(-1, -1, -1)

def test_vector_pos():
    assert +Vector(1, 1, 1) == Vector(1, 1, 1)

def test_vector_abs():
    assert abs(Vector(-1, 1, -2)) == Vector(1, 1, 2)
    assert abs(Vector(0, 1, 2)) == Vector(0, 1, 2)

def test_vector_bool():
    assert Vector(1, 0, 0)
    assert Vector(0, 1, 0)
    assert Vector(0, 0, 1)
    assert not Vector()

def test_vector_dot():
    assert Vector(1, 1, 1).dot(Vector()) == 0
    assert Vector(1, 2, 3).dot(Vector(1, 1, 1)) == 6
    assert Vector(1, 2, 3).dot(Vector(0, 0, 2)) == 6

def test_vector_cross():
    assert Vector(x=1).cross(Vector(x=1)) == Vector()
    assert Vector(y=1).cross(Vector(y=-1)) == Vector()
    assert Vector(z=1).cross(Vector(z=1)) == Vector()
    assert Vector(x=1).cross(Vector(y=1)) == Vector(z=1)
    assert Vector(x=1).cross(Vector(y=-1)) == Vector(z=-1)

def test_vector_distance_to():
    assert Vector().distance_to(Vector()) == 0.0
    assert Vector(x=1).distance_to(Vector(x=2)) == 1.0
    assert Vector(x=1).distance_to(Vector(2, 1, 0)) == 2 ** 0.5

def test_vector_magnitude():
    assert Vector().magnitude == 0
    assert Vector(2, 4, 4).magnitude == 6

def test_vector_unit():
    assert Vector(x=1).unit == Vector(1, 0, 0)
    assert Vector().unit == Vector()
    assert Vector(2, 4, 4).unit == Vector(1/3, 2/3, 2/3)

def test_vector_replace():
    assert Vector().replace(x=1) == Vector(1, 0, 0)
    assert Vector(1, 2, 3).replace() == Vector(1, 2, 3)
    assert Vector(1, 2, 3).replace(2, 4, 6) == Vector(2, 4, 6)
    assert Vector(1, 2, 3).replace(z=-1) == Vector(1, 2, -1)

def test_vector_trunc():
    assert math.trunc(Vector(1, 2, 3)) == Vector(1, 2, 3)
    assert math.trunc(Vector(1.1, 2.5, -1.1)) == Vector(1, 2, -1)
    assert math.trunc(Vector(1.9, 0.0, -1.9)) == Vector(1, 0, -1)

def test_vector_floor():
    assert Vector(1, 2, 3).floor() == Vector(1, 2, 3)
    assert Vector(1.1, 2.5, -1.1).floor() == Vector(1, 2, -2)
    assert Vector(1.9, 0.0, -1.9).floor() == Vector(1, 0, -2)

def test_vector_ceil():
    assert Vector(1, 2, 3).ceil() == Vector(1, 2, 3)
    assert Vector(1.1, 2.5, -1.1).ceil() == Vector(2, 3, -1)
    assert Vector(1.9, 0.0, -1.9).ceil() == Vector(2, 0, -1)

def test_vector_round():
    assert fp_vectors_equal(Vector(1, 2, 3).round(), Vector(1, 2, 3))
    assert fp_vectors_equal(Vector(1.1, 3.5, -1.1).round(), Vector(1, 4, -1))
    assert fp_vectors_equal(Vector(1.9, 0.0, -1.9).round(), Vector(2, 0, -2))
    assert fp_vectors_equal(Vector(1.9, 0.0, -1.9).round(1), Vector(1.9, 0.0, -1.9))
    assert fp_vectors_equal(Vector(1.9, 0.0, -1.9).round(-1), Vector(0, 0, 0))

def test_vector_angle_between():
    assert fp_equal(X.angle_between(Y), 90.0)
    assert fp_equal(Y.angle_between(Z), 90.0)
    assert fp_equal((X + Y).angle_between(X), 45.0)

def test_vector_project():
    assert fp_equal(X.project(X), 1.0)
    assert fp_equal(X.project(Y), 0.0)
    assert fp_equal(X.project(Z), 0.0)
    assert fp_equal(Vector(1, 2, 3).project(2 * Y), 2.0)
    assert fp_equal(Vector(3, 4, 5).project(Vector(3, 4, 0)), 5.0)

def test_vector_rotate():
    assert fp_vectors_equal(X.rotate(90, about=X), X)
    assert fp_vectors_equal(X.rotate(90, about=Y), -Z)
    assert fp_vectors_equal(X.rotate(90, about=Z), Y)
    assert fp_vectors_equal(X.rotate(90, about=-X), X)
    assert fp_vectors_equal(X.rotate(90, about=-Y), Z)
    assert fp_vectors_equal(X.rotate(90, about=-Z), -Y)
    assert fp_vectors_equal(X.rotate(180, about=X + Y), Y)
    assert fp_vectors_equal(O.rotate(180, about=Y, origin=Z), 2*Z)

def test_vector_range_init():
    v = vector_range(Vector() + 2)
    assert v.start == Vector()
    assert v.stop == Vector(2, 2, 2)
    assert v.step == Vector(1, 1, 1)
    with pytest.raises(TypeError):
        vector_range(Vector(0.0, 0.0, 0.0), step=Vector(0.5, 0.5, 1.0))

def test_vector_range_start():
    assert vector_range(Vector() + 1).start == Vector()
    assert vector_range(Vector(), Vector() + 1).start == Vector()
    assert vector_range(Vector(1, 0, 1), Vector() + 1).start == Vector(1, 0, 1)

def test_vector_range_stop():
    assert vector_range(Vector() + 1).stop == Vector(1, 1, 1)
    assert vector_range(Vector(1, 1, 0)).stop == Vector(1, 1, 0)
    assert vector_range(Vector(1, 0, 1), Vector() + 1).stop == Vector(1, 1, 1)

def test_vector_range_step():
    assert vector_range(Vector() + 1, step=Vector() + 2).step == Vector(2, 2, 2)
    assert vector_range(Vector(), Vector() + 1, Vector() + 1).step == Vector(1, 1, 1)
    with pytest.raises(ValueError):
        vector_range(Vector() + 1, step=Vector(1, 1, 0))

def test_vector_range_order():
    assert vector_range(Vector() + 1).order == 'zxy'
    assert vector_range(Vector() + 1, order='xyz').order == 'xyz'
    with pytest.raises(ValueError):
        vector_range(Vector() + 2, order='abc')

def test_vector_range_index():
    assert vector_range(Vector() + 2).index(Vector()) == 0
    assert vector_range(Vector() + 2)[:Vector() - 1].index(Vector()) == 0
    assert vector_range(Vector() + 2).index(Vector(1, 1, 1)) == 7
    assert vector_range(Vector() + 2)[Vector(z=1):].index(Vector(1, 1, 1)) == 3
    assert vector_range(Vector() + 2, order='xyz').index(Vector(1, 0, 0)) == 1
    assert vector_range(Vector() + 2, order='zxy').index(Vector(1, 0, 0)) == 2
    assert vector_range(Vector() + 2, order='zyx').index(Vector(1, 0, 0)) == 4
    with pytest.raises(ValueError):
        vector_range(Vector() + 2).index(Vector(2, 2, 2))

def test_vector_range_contains():
    assert Vector() in vector_range(Vector() + 2)
    assert Vector(1, 1, 1) in vector_range(Vector() + 2)
    assert Vector(x=1) in vector_range(Vector() + 2, order='xyz')
    assert Vector() + 2 not in vector_range(Vector() + 2)

def test_vector_range_count():
    assert vector_range(Vector() + 2).count(Vector()) == 1
    assert vector_range(Vector() + 2).count(Vector(1, 1, 1)) == 1
    assert vector_range(Vector() + 2, order='xyz').count(Vector(1, 0, 0)) == 1
    assert vector_range(Vector() + 2).count(Vector(2, 2, 2)) == 0

def test_vector_range_len():
    assert len(vector_range(Vector())) == 0
    assert len(vector_range(Vector() + 1)) == 1
    assert len(vector_range(Vector() + 2)) == 8
    assert len(vector_range(Vector() + 3)) == 27

def test_vector_range_ordering():
    assert vector_range(Vector() + 2, order='xyz') == vector_range(Vector(2, 2, 2), order='xyz')
    assert vector_range(Vector() + 2, order='xyz') == list(vector_range(Vector(2, 2, 2), order='xyz'))

    assert vector_range(Vector() + 2, order='zxy') <= vector_range(Vector() + 2, order='xyz')
    assert vector_range(Vector() + 2, order='zxy') <= vector_range(Vector() + 2, order='zxy')
    assert vector_range(Vector() + 2, order='zxy') < vector_range(Vector() + 2, order='xyz')
    assert not (vector_range(Vector() + 2, order='zxy') < vector_range(Vector() + 2, order='zxy'))

    assert vector_range(Vector() + 2, order='xyz') > vector_range(Vector() + 2, order='zxy')
    assert vector_range(Vector() + 2, order='xyz') >= vector_range(Vector() + 2, order='zxy')
    assert vector_range(Vector() + 2, order='xyz') != vector_range(Vector() + 2, order='zxy')

    assert vector_range(Vector() + 2, order='xyz') == [
            Vector(0, 0, 0), Vector(1, 0, 0), Vector(0, 1, 0), Vector(1, 1, 0),
            Vector(0, 0, 1), Vector(1, 0, 1), Vector(0, 1, 1), Vector(1, 1, 1),
            ]
    assert vector_range(Vector() + 2, order='zxy') == [
            Vector(0, 0, 0), Vector(0, 0, 1), Vector(1, 0, 0), Vector(1, 0, 1),
            Vector(0, 1, 0), Vector(0, 1, 1), Vector(1, 1, 0), Vector(1, 1, 1),
            ]
    assert vector_range(Vector() + 2, order='zyx') == [
            Vector(0, 0, 0), Vector(0, 0, 1), Vector(0, 1, 0), Vector(0, 1, 1),
            Vector(1, 0, 0), Vector(1, 0, 1), Vector(1, 1, 0), Vector(1, 1, 1),
            ]

    assert iter(vector_range(Vector() + 2, order='zxy')) < vector_range(Vector() + 2, order='xyz')
    assert iter(vector_range(Vector() + 2, order='zxy')) <= vector_range(Vector() + 2, order='xyz')
    assert iter(vector_range(Vector() + 2, order='xyz')) > vector_range(Vector() + 2, order='zxy')
    assert iter(vector_range(Vector() + 2, order='xyz')) >= vector_range(Vector() + 2, order='zxy')
    assert iter(vector_range(Vector() + 2, order='xyz')) != vector_range(Vector() + 2, order='zxy')

def test_vector_range_reversed():
    assert list(reversed(vector_range(Vector() + 2))) == list(reversed(list(vector_range(Vector() + 2))))
    assert list(reversed(vector_range(Vector() + 2, order='xyz'))) == vector_range(Vector() + 1, Vector() - 1, Vector() - 1, order='xyz')
    assert list(reversed(vector_range(Vector() + 2, order='xyz'))) == vector_range(Vector() + 2, order='xyz')[::Vector() - 1]

def test_vector_range_bool():
    assert not vector_range(Vector())
    assert vector_range(Vector() + 1)
    assert not vector_range(Vector() + 1)[Vector() + 1:]

def test_vector_range_getitem():
    v = vector_range(Vector() + 2, order='xyz')
    assert v[0] == Vector()
    assert v[1] == Vector(1, 0, 0)
    assert v[7] == Vector(1, 1, 1)
    assert v[7] == v[-1]
    assert v[6] == v[-2]
    assert v[0] == v[-8]
    assert v[Vector()] == Vector()
    assert v[Vector(1, 0, 0)] == Vector(1, 0, 0)
    assert v[Vector() - 1] == Vector(1, 1, 1)
    with pytest.raises(IndexError):
        v[8]
    with pytest.raises(IndexError):
        v[-9]
    with pytest.raises(IndexError):
        v[Vector(2, 1, 1)]

def test_vector_getslice():
    v = vector_range(Vector() + 2, order='xyz')
    assert v[Vector(1, 0, 0):] == vector_range(Vector(x=1), Vector() + 2, order='xyz')
    assert v[:Vector(1, None, None)] == vector_range(Vector(1, 2, 2), order='xyz')
    with pytest.raises(ValueError):
        v[1:]
    with pytest.raises(ValueError):
        v[::Vector(1, 1, 0)]

def test_vector_coverage():
    # Miscellaneous tests purely for the sake of coverage
    assert rmod(3, 3, range(10)) == set()
    assert rmod(3, 2, []) == set()
    with pytest.raises(ValueError):
        rmod(0, 1, range(10))
    with pytest.raises(ValueError):
        rdiv(0, 1)

def test_vector_sign():
    assert sign(10) == 1
    assert sign(0) == 0
    assert sign(-5) == -1
    assert sign(X) == X
    assert sign(O) == O
    assert sign(Vector(3, 0, -3)) == X - Z

def test_vector_line():
    assert list(line(O, O)) == [O]
    assert list(line(O, 4*X)) == [O, X, 2*X, 3*X, 4*X]
    assert list(line(O, Vector(3, 1, 0))) == [O, X, 2*X + Y, 3*X + Y]
    assert list(line(O, Vector(1, 2, 3))) == [O, Y + Z, Vector(1, 1, 2), Vector(1, 2, 3)]

def test_vector_lines():
    assert list(lines([O, 4*X, Vector(1, 2, 3)])) == [
            O, X, 2*X, 3*X, 4*X, Vector(3, 1, 1), Vector(2, 1, 2),
            Vector(1, 2, 3), Vector(1, 1, 2), Vector(0, 1, 1), O]
    assert list(lines([O, 4*X, Vector(1, 2, 3)], closed=False)) == [
            O, X, 2*X, 3*X, 4*X, Vector(3, 1, 1), Vector(2, 1, 2),
            Vector(1, 2, 3)]




# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# An alternate Python Minecraft library for the Rasperry-Pi
# Copyright (c) 2013-2015 Dave Jones <dave@waveform.org.uk>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
The world module defines the :class:`World` class, which is the usual way of
starting a connection to a Minecraft server and which then provides various
attributes allowing the user to query and manipulate that world.

.. note::

    All items in this module are available from the :mod:`picraft` namespace
    without having to import :mod:`picraft.world` directly.

The following items are defined in the module:


World
=====

.. autoclass:: World
    :members:


Checkpoint
==========

.. autoclass:: Checkpoint
    :members:


Camera
======

.. autoclass:: Camera
    :members:
"""

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
str = type('')


from .exc import NotSupported
from .connection import Connection
from .player import HostPlayer, Players
from .block import Blocks
from .vector import Vector, vector_range
from .events import Events


class World(object):
    """
    Represents a Minecraft world.

    This is the primary class that users interact with. Construct an instance
    of this class, optionally specifying the *host* and *port* of the server
    (which default to "localhost" and 4711 respectively). Afterward, the
    instance can be used to query and manipulate the minecraft world of the
    connected game.

    The :meth:`say` method can be used to send commands to the console, while
    the :attr:`player` attribute can be used to manipulate or query the status
    of the player character in the world. The :attr:`players` attribute can be
    used to manipulate or query other players within the world (this object can
    be iterated over to discover players)::

        >>> from picraft import *
        >>> world = World()
        >>> len(world.players)
        1
        >>> world.say('Hello, world!')
    """

    def __init__(
            self, host='localhost', port=4711, timeout=0.3,
            ignore_errors=False):
        self._connection = Connection(host, port, timeout, ignore_errors)
        self._player = HostPlayer(self._connection)
        self._players = Players(self._connection)
        self._blocks = Blocks(self._connection)
        self._height = WorldHeight(self._connection)
        self._checkpoint = Checkpoint(self._connection)
        self._camera = Camera(self._connection)
        self._events = Events(self._connection)

    def __repr__(self):
        return '<World players=%d>' % len(self.players)

    @property
    def connection(self):
        """
        Represents the connection to the Minecraft server.

        The :class:`~picraft.connection.Connection` object contained in this
        attribute represents the connection to the Minecraft server and
        provides various methods for communicating with it. Users will very
        rarely need to access this attribute, except to use the
        :meth:`~picraft.connection.Connection.batch_start` method.
        """
        return self._connection

    @property
    def players(self):
        """
        Represents all player entities in the Minecraft world.

        This property can be queried to determine which players are currently
        in the Minecraft world. The property is a mapping of player id (an
        integer number) to a :class:`~picraft.player.Player` object which
        permits querying and manipulation of the player. The property supports
        many of the methods of dicts and can be iterated over like a dict::

            >>> len(world.players)
            1
            >>> list(world.players)
            [1]
            >>> world.players.keys()
            [1]
            >>> world.players[1]
            <picraft.player.Player at 0x7f2f91f38cd0>
            >>> world.players.values()
            [<picraft.player.Player at 0x7f2f91f38cd0>]
            >>> world.players.items()
            [(1, <picraft.player.Player at 0x7f2f91f38cd0>)]
            >>> for player in world.players:
            ...     print(player.tile_pos)
            ...
            -3,18,-5

        On the Raspberry Juice platform, you can also use player name to
        reference players::

            >>> world.players['my_player']
            <picraft.player.Player at 0x7f2f91f38cd0>
        """
        return self._players

    @property
    def player(self):
        """
        Represents the host player in the Minecraft world.

        The :class:`~picraft.player.HostPlayer` object returned by this
        attribute provides properties which can be used to query the status of,
        and manipulate the state of, the host player in the Minecraft world::

            >>> world.player.pos
            Vector(x=-2.49725, y=18.0, z=-4.21989)
            >>> world.player.tile_pos += Vector(y=50)
        """
        return self._player

    @property
    def height(self):
        """
        Represents the height of the Minecraft world.

        This property can be queried to determine the height of the world at
        any location. The property can be indexed with a single
        :class:`~picraft.vector.Vector`, in which case the height will be
        returned as a vector with the same X and Z coordinates, but a Y
        coordinate adjusted to the first non-air block from the top of the
        world::

            >>> world.height[Vector(0, -10, 0)]
            Vector(x=0, y=0, z=0)

        Alternatively, a slice of two vectors can be used. In this case, the
        property returns a sequence of :class:`~picraft.vector.Vector` objects
        each with their Y coordinates adjusted to the height of the world at
        the respective X and Z coordinates.
        """
        return self._height

    @property
    def camera(self):
        """
        Represents the camera of the Minecraft world.

        The :class:`Camera` object contained in this property permits control
        of the position of the virtual camera in the Minecraft world. For
        example, to position the camera directly above the host player::

            >>> world.camera.third_person(world.player)

        Alternatively, to see through the eyes of a specific player::

            >>> world.camera.first_person(world.players[2])

        .. warning::

            Camera control is only supported on Minecraft Pi edition.
        """
        return self._camera

    @property
    def blocks(self):
        """
        Represents the state of blocks in the Minecraft world.

        This property can be queried to determine the type of a block in the
        world, or can be set to alter the type of a block. The property can be
        indexed with a single :class:`~picraft.vector.Vector`, in which case
        the state of a single block is returned (or updated) as a
        :class:`~picraft.block.Block` object::

            >>> world.blocks[g.player.tile_pos]
            <Block "grass" id=2 data=0>

        Alternatively, a slice of vectors can be used. In this case, when
        querying the property, a sequence of :class:`~picraft.block.Block`
        objects is returned, When setting a slice of vectors you can either
        pass a sequence of :class:`~picraft.block.Block` objects or a single
        :class:`~picraft.block.Block` object. The sequence must be equal to
        the number of blocks represented by the slice::

            >>> world.blocks[Vector(0,0,0):Vector(2,1,1)]
            [<Block "grass" id=2 data=0>,<Block "grass" id=2 data=0>]
            >>> world.blocks[Vector(0,0,0):Vector(5,1,5)] = Block.from_name('grass')

        As with normal Python slices, the interval specified is `half-open`_.
        That is to say, it is inclusive of the lower vector, *exclusive* of the
        upper one. Hence, ``Vector():Vector(x=5,1,1)`` represents the
        coordinates (0,0,0) to (4,0,0). It is usually useful to specify the
        upper bound as the vector you want and then add one to it::

            >>> world.blocks[Vector():Vector(x=1) + 1]
            [<Block "grass" id=2 data=0>,<Block "grass" id=2 data=0>]
            >>> world.blocks[Vector():Vector(4,0,4) + 1] = Block.from_name('grass')

        .. _half-open: http://python-history.blogspot.co.uk/2013/10/why-python-uses-0-based-indexing.html

        .. warning::

            Querying or setting sequences of blocks can be extremely slow as a
            network transaction must be executed for each individual block.
            When setting a slice of blocks, this can be speeded up by
            specifying a single :class:`~picraft.block.Block` in which case one
            network transaction will occur to set all blocks in the slice.  The
            Raspberry Juice server also supports querying sequences of blocks
            with a single command (picraft will automatically use this).
            Additionally, :meth:`~picraft.connection.Connection.batch_start`
            can be used to speed up setting sequences of blocks (though not
            querying).
        """
        return self._blocks

    @property
    def events(self):
        """
        Provides an interface to poll events that occur in the Minecraft world.

        The :class:`~picraft.events.Events` object contained in this property
        provides methods for determining what is happening in the Minecraft
        world::

            >>> events = world.events.poll()
            >>> len(events)
            3
            >>> events[0]
            <BlockHitEvent pos=1,1,1 face="x+" player=1>
            >>> events[0].player.pos
            <Vector x=0.5, y=0.0, z=0.5>
        """
        return self._events

    @property
    def checkpoint(self):
        """
        Represents the Minecraft world checkpoint system.

        The :class:`Checkpoint` object contained in this attribute provides the
        ability to save and restore the state of the world at any time::

            >>> world.checkpoint.save()
            >>> world.blocks[Vector()] = Block.from_name('stone')
            >>> world.checkpoint.restore()
        """
        return self._checkpoint

    def say(self, message):
        """
        Displays *message* in the game's chat console.

        The *message* parameter must be a string (which may contain multiple
        lines). Each line of the message will be sent to the game's chat
        console and displayed immediately. For example::

            >>> world.say('Hello, world!')
            >>> world.say('The following player IDs exist:\\n%s' %
            ...     '\\n'.join(str(p) for p in world.players))
        """
        for line in message.splitlines():
            self.connection.send('chat.post(%s)' % line)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.connection.close()

    def _get_immutable(self):
        raise AttributeError(
                'reading immutable is not supported by the server')
    def _set_immutable(self, value):
        if self._connection.server_version != 'minecraft-pi':
            raise NotSupported(
                'cannot change world settings on server version: %s' %
                self._connection.server_version)
        self._connection.send('world.setting(world_immutable,%d)' % bool(value))
    immutable = property(_get_immutable, _set_immutable,
        doc="""\
        Write-only property which sets whether the world is changeable.

        .. warning::

            World settings are only supported on Minecraft Pi edition.

        .. note::

            Unfortunately, the underlying protocol provides no means of reading
            a world setting, so this property is write-only (attempting to
            query it will result in an :exc:`AttributeError` being raised).
        """)

    def _get_nametags_visible(self):
        raise AttributeError(
                'reading nametags_visible is not supported by the server')
    def _set_nametags_visible(self, value):
        if self._connection.server_version != 'minecraft-pi':
            raise NotSupported(
                'cannot change world settings on server version: %s' %
                self._connection.server_version)
        self._connection.send('world.setting(nametags_visible,%d)' % bool(value))
    nametags_visible = property(_get_nametags_visible, _set_nametags_visible,
        doc="""\
        Write-only property which sets whether players' nametags are visible.

        .. warning::

            World settings are only supported on Minecraft Pi edition.

        .. note::

            Unfortunately, the underlying protocol provides no means of reading
            a world setting, so this property is write-only (attempting to
            query it will result in an :exc:`AttributeError` being raised).
        """)


class WorldHeight(object):
    """
    This class implements the :attr:`~picraft.world.World.heights` attribute.
    """

    def __init__(self, connection):
        self._connection = connection

    def __repr__(self):
        return '<WorldHeight>'

    def __getitem__(self, index):
        if isinstance(index, slice):
            return [
                Vector(v.x, int(self._connection.transact(
                    'world.getHeight(%d,%d)' % (v.x, v.z))), v.z)
                for v in vector_range(index.start, index.stop)
                ]
        else:
            return Vector(index.x, int(self._connection.transact(
                'world.getHeight(%d,%d)' % (index.x, index.z))), index.z)


class Checkpoint(object):
    """
    Permits restoring the world state from a prior save.

    This class provides methods for storing the state of the Minecraft world,
    and restoring the saved state at a later time. The :meth:`save` method
    saves the state of the world, and the :meth:`restore` method restores
    the saved state.

    This class can be used as a context manager to take a checkpoint, make
    modifications to the world, and roll them back if an exception occurs.
    For example, the following code will ultimately do nothing because an
    exception occurs after the alteration::

        >>> from picraft import *
        >>> w = World()
        >>> with w.checkpoint:
        ...     w.blocks[w.player.tile_pos - Vector(y=1)] = Block.from_name('stone')
        ...     raise Exception()

    .. warning::

        Checkpoints are only supported on Minecraft Pi edition.

    .. warning::

        Minecraft only permits a single checkpoint to be stored at any given
        time. There is no capability to save multiple checkpoints and no way of
        checking whether one currently exists. Therefore, storing a checkpoint
        may overwrite an older checkpoint without warning.

    .. note::
        Checkpoints don't work *within* batches as the checkpoint save will be
        batched along with everything else. That said, a checkpoint can be used
        *outside* a batch to roll the entire thing back if it fails::

            >>> v = w.player.tile_pos - Vector(y=1)
            >>> with w.checkpoint:
            ...     with w.connection.batch_start():
            ...         w.blocks[v - Vector(2, 0, 2):v + Vector(2, 1, 2)] = [
            ...             Block.from_name('wool', data=i) for i in range(16)]
    """

    def __init__(self, connection):
        self._connection = connection

    def __repr__(self):
        return '<Checkpoint>'

    def save(self):
        """
        Save the state of the Minecraft world, overwriting any prior checkpoint
        state.
        """
        if self._connection.server_version != 'minecraft-pi':
            raise NotSupported(
                'cannot save checkpoint on server version: %s' %
                self._connection.server_version)
        self._connection.send('world.checkpoint.save()')

    def restore(self):
        """
        Restore the state of the Minecraft world from a previously saved
        checkpoint.  No facility is provided to determine whether a prior
        checkpoint is available (the underlying network protocol doesn't permit
        this).
        """
        if self._connection.server_version != 'minecraft-pi':
            raise NotSupported(
                'cannot restore checkpoint on server version: %s' %
                self._connection.server_version)
        self._connection.send('world.checkpoint.restore()')

    def __enter__(self):
        self.save()

    def __exit__(self, exc_type, exc_value, exc_tb):
        if exc_type is not None:
            self.restore()


class Camera(object):
    """
    This class implements the :attr:`~picraft.world.World.camera` attribute.
    """

    def __init__(self, connection):
        self._connection = connection

    def __repr__(self):
        return '<Camera>'

    def _get_pos(self):
        raise AttributeError(
                'reading camera position is not supported by server')
    def _set_pos(self, value):
        if self._connection.server_version != 'minecraft-pi':
            raise NotSupported(
                'cannot position camera on server version: %s' %
                    self._connection.server_version)
        self._connection.send('camera.mode.setFixed()')
        self._connection.send('camera.setPos(%d,%d,%d)' % (value.x, value.y, value.z))
    pos = property(_get_pos, _set_pos, doc="""\
        Write-only property which sets the camera's absolute position in the
        world.

        .. note::

            Unfortunately, the underlying protocol provides no means of reading
            this setting, so this property is write-only (attempting to query
            it will result in an :exc:`AttributeError` being raised).
        """)

    def third_person(self, player):
        """
        Causes the camera to follow the specified player from above. The
        *player* can be the :attr:`~World.player` attribute (representing the
        host player) or an attribute retrieved from the :attr:`~World.players`
        list. For example::

            >>> from picraft import World
            >>> w = World()
            >>> w.camera.third_person(w.player)
            >>> w.camera.third_person(w.players[1])
        """
        if self._connection.server_version != 'minecraft-pi':
            raise NotSupported(
                'cannot position camera on server version: %s' %
                self._connection.server_version)
        if isinstance(player, HostPlayer):
            self._connection.send('camera.mode.setFollow()')
        else:
            self._connection.send('camera.mode.setFollow(%d)' % player.player_id)

    def first_person(self, player):
        """
        Causes the camera to view the world through the eyes of the specified
        player. The *player* can be the :attr:`~World.player` attribute
        (representing the host player) or an attribute retrieved from the
        :attr:`~World.players` list. For example::

            >>> from picraft import World
            >>> w = World()
            >>> w.camera.first_person(w.player)
            >>> w.camera.first_person(w.players[1])
        """
        if self._connection.server_version != 'minecraft-pi':
            raise NotSupported(
                'cannot position camera on server version: %s' %
                self._connection.server_version)
        if isinstance(player, HostPlayer):
            self._connection.send('camera.mode.setNormal()')
        else:
            self._connection.send('camera.mode.setNormal(%d)' % player.player_id)




# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# An alternate Python Minecraft library for the Rasperry-Pi
# Copyright (c) 2013-2015 Dave Jones <dave@waveform.org.uk>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
str = type('')


import pytest
import io
import picraft.events
from picraft import World, Vector
try:
    from unittest import mock
except ImportError:
    import mock


def test_events_poll_empty():
    conn = mock.MagicMock()
    conn.transact.return_value = ''
    events = picraft.events.Events(conn)
    assert events.poll() == []
    conn.transact.assert_called_once_with('events.block.hits()')

def test_events_poll_one_hit():
    conn = mock.MagicMock()
    conn.transact.return_value = '1,2,3,4,5'
    events = picraft.events.Events(conn)
    result = events.poll()
    assert len(result) == 1
    assert result[0].pos == Vector(1, 2, 3)
    assert result[0].face == 'x-'
    assert result[0].player.player_id == 5
    conn.transact.assert_called_once_with('events.block.hits()')

def test_events_poll_multi_hits():
    conn = mock.MagicMock()
    conn.transact.return_value = '1,2,3,4,5|-1,0,0,0,1'
    events = picraft.events.Events(conn)
    result = events.poll()
    assert len(result) == 2
    assert result[0].pos == Vector(1, 2, 3)
    assert result[0].face == 'x-'
    assert result[0].player.player_id == 5
    assert result[1].pos == Vector(-1, 0, 0)
    assert result[1].face == 'y-'
    assert result[1].player.player_id == 1
    conn.transact.assert_called_once_with('events.block.hits()')

def test_events_clear():
    conn = mock.MagicMock()
    events = picraft.events.Events(conn)
    events.clear()
    conn.send.assert_called_once_with('events.clear()')




# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# An alternate Python Minecraft library for the Rasperry-Pi
# Copyright (c) 2013-2015 Dave Jones <dave@waveform.org.uk>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
The vector module defines the :class:`Vector` class, which is the usual method
of represent coordinates or vectors when dealing with the Minecraft world. It
also provides functions like :func:`vector_range` for generating sequences of
vectors.

.. note::

    All items in this module are available from the :mod:`picraft` namespace
    without having to import :mod:`picraft.vector` directly.

The following items are defined in the module:


Vector
======

.. autoclass:: Vector(x=0, y=0, z=0)


Short-hand variants
===================

The :class:`Vector` class is used sufficiently often to justify the inclusion
of some shortcuts. The class itself is also available as ``V``, and vectors
representing the three axes are each available as ``X``, ``Y``, and ``Z``.
Finally, a vector representing the origin is available as ``O``::

    >>> from picraft import V, O, X, Y, Z
    >>> O
    Vector(x=0, y=0, z=0)
    >>> 2 * X
    Vector(x=2, y=0, z=0)
    >>> X + Y
    Vector(x=1, y=1, z=0)
    >>> (X + Y).angle_between(X)
    45.00000000000001
    >>> V(3, 4, 5).projection(X)
    3.0
    >>> X.rotate(90, about=Y)
    Vector(x=0.0, y=0.0, z=1.0)


vector_range
============

.. autoclass:: vector_range
    :members:


line
====

.. autofunction:: line


lines
=====

.. autofunction:: lines
"""

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
str = type('')
from .compat import range


import math
from functools import total_ordering
from collections import namedtuple, Sequence
try:
    from itertools import zip_longest, islice
except ImportError:
    # Py2 compat
    from itertools import izip_longest as zip_longest, islice


class Vector(namedtuple('Vector', ('x', 'y', 'z'))):
    """
    Represents a 3-dimensional vector.

    This tuple derivative represents a 3-dimensional vector with x, y, z
    components. Instances can be constructed in a number of ways. By explicitly
    specifying the x, y, and z components (optionally with keyword
    identifiers), or leaving the empty to default to 0::

        >>> Vector(1, 1, 1)
        Vector(x=1, y=1, z=1)
        >>> Vector(x=2, y=0, z=0)
        Vector(x=2, y=0, z=0)
        >>> Vector()
        Vector(x=0, y=0, z=0)
        >>> Vector(y=10)
        Vector(x=0, y=10, z=0)

    Shortcuts are available for the X, Y, and Z axes::

        >>> X
        Vector(x=1, y=0, z=0)
        >>> Y
        Vector(x=0, y=1, z=0)

    Note that vectors don't much care whether their components are integers,
    floating point values, or ``None``::

        >>> Vector(1.0, 1, 1)
        Vector(x=1.0, y=1, z=1)
        >>> Vector(2, None, None)
        Vector(x=2, y=None, z=None)

    The class supports simple arithmetic operations with other vectors such as
    addition and subtraction, along with multiplication and division with
    scalars, raising to powers, bit-shifting, and so on. Such operations are
    performed element-wise [1]_::

        >>> v1 = Vector(1, 1, 1)
        >>> v2 = Vector(2, 2, 2)
        >>> v1 + v2
        Vector(x=3, y=3, z=3)
        >>> 2 * v2
        Vector(x=4, y=4, z=4)

    Simple arithmetic operations with scalars return a new vector with that
    operation performed on all elements of the original. For example::

        >>> v = Vector()
        >>> v
        Vector(x=0, y=0, z=0)
        >>> v + 1
        Vector(x=1, y=1, z=1)
        >>> 2 * (v + 2)
        Vector(x=4, y=4, z=4)
        >>> Vector(y=2) ** 2
        Vector(x=0, y=4, z=0)

    Within the Minecraft world, the X,Z plane represents the ground, while the
    Y vector represents height.

    .. note::

        Note that, as a derivative of tuple, instances of this class are
        immutable. That is, you cannot directly manipulate the x, y, and z
        attributes; instead you must create a new vector (for example, by
        adding two vectors together). The advantage of this is that vector
        instances can be used in sets or as dictionary keys.

    .. [1] I realize math purists will hate this (and demand that abs() should
       be magnitude and * should invoke matrix multiplication), but the
       element wise operations are sufficiently useful to warrant the
       short-hand syntax.

    .. automethod:: replace

    .. automethod:: ceil

    .. automethod:: floor

    .. automethod:: dot

    .. automethod:: cross

    .. automethod:: distance_to

    .. automethod:: angle_between

    .. automethod:: project

    .. automethod:: rotate

    .. attribute:: x

        The position or length of the vector along the X-axis. In the Minecraft
        world this can be considered to run left-to-right.

    .. attribute:: y

        The position or length of the vector along the Y-axis. In the Minecraft
        world this can be considered to run vertically up and down.

    .. attribute:: z

        The position or length of the vector along the Z-axis. In the Minecraft
        world this can be considered as depth (in or out of the screen).

    .. autoattribute:: magnitude

    .. autoattribute:: unit
    """

    def __new__(cls, x=0, y=0, z=0):
        return super(Vector, cls).__new__(cls, x, y, z)

    @classmethod
    def from_string(cls, s, type=int):
        x, y, z = s.split(',', 2)
        return cls(type(x), type(y), type(z))

    @property
    def __dict__(self):
        # This is required to work around a subtle issue encountered in Python
        # 3.3 and above. In these versions (probably deliberately), the
        # __dict__ property is not inherited by namedtuple descendents
        return super(Vector, self).__dict__

    def __str__(self):
        return '%s,%s,%s' % (self.x, self.y, self.z)

    def __add__(self, other):
        try:
            return Vector(self.x + other.x, self.y + other.y, self.z + other.z)
        except AttributeError:
            return Vector(self.x + other, self.y + other, self.z + other)

    __radd__ = __add__

    def __sub__(self, other):
        try:
            return Vector(self.x - other.x, self.y - other.y, self.z - other.z)
        except AttributeError:
            return Vector(self.x - other, self.y - other, self.z - other)

    def __mul__(self, other):
        try:
            return Vector(self.x * other.x, self.y * other.y, self.z * other.z)
        except AttributeError:
            return Vector(self.x * other, self.y * other, self.z * other)

    __rmul__ = __mul__

    def __truediv__(self, other):
        try:
            return Vector(self.x / other.x, self.y / other.y, self.z / other.z)
        except AttributeError:
            return Vector(self.x / other, self.y / other, self.z / other)

    def __floordiv__(self, other):
        try:
            return Vector(self.x // other.x, self.y // other.y, self.z // other.z)
        except AttributeError:
            return Vector(self.x // other, self.y // other, self.z // other)

    def __mod__(self, other):
        try:
            return Vector(self.x % other.x, self.y % other.y, self.z % other.z)
        except AttributeError:
            return Vector(self.x % other, self.y % other, self.z % other)

    def __pow__(self, other, modulo=None):
        if modulo is not None:
            try:
                # XXX What about other vector, modulo scalar, and other scalar, modulo vector?
                return Vector(
                        pow(self.x, other.x, modulo.x),
                        pow(self.y, other.y, modulo.y),
                        pow(self.z, other.z, modulo.z))
            except AttributeError:
                return Vector(
                        pow(self.x, other, modulo),
                        pow(self.y, other, modulo),
                        pow(self.z, other, modulo))
        try:
            return Vector(
                    pow(self.x, other.x),
                    pow(self.y, other.y),
                    pow(self.z, other.z))
        except AttributeError:
            return Vector(
                    pow(self.x, other),
                    pow(self.y, other),
                    pow(self.z, other))

    def __lshift__(self, other):
        try:
            return Vector(self.x << other.x, self.y << other.y, self.z << other.z)
        except AttributeError:
            return Vector(self.x << other, self.y << other, self.z << other)

    def __rshift__(self, other):
        try:
            return Vector(self.x >> other.x, self.y >> other.y, self.z >> other.z)
        except AttributeError:
            return Vector(self.x >> other, self.y >> other, self.z >> other)

    def __and__(self, other):
        try:
            return Vector(self.x & other.x, self.y & other.y, self.z & other.z)
        except AttributeError:
            return Vector(self.x & other, self.y & other, self.z & other)

    def __xor__(self, other):
        try:
            return Vector(self.x ^ other.x, self.y ^ other.y, self.z ^ other.z)
        except AttributeError:
            return Vector(self.x ^ other, self.y ^ other, self.z ^ other)

    def __or__(self, other):
        try:
            return Vector(self.x | other.x, self.y | other.y, self.z | other.z)
        except AttributeError:
            return Vector(self.x | other, self.y | other, self.z | other)

    def __neg__(self):
        return Vector(-self.x, -self.y, -self.z)

    def __pos__(self):
        return self

    def __abs__(self):
        return Vector(abs(self.x), abs(self.y), abs(self.z))

    def __bool__(self):
        return bool(self.x or self.y or self.z)

    def __trunc__(self):
        return Vector(math.trunc(self.x), math.trunc(self.y), math.trunc(self.z))

    # Py2 compat
    __nonzero__ = __bool__
    __div__ = __truediv__

    def replace(self, x=None, y=None, z=None):
        """
        Return the vector with the x, y, or z axes replaced with the specified
        values. For example::

            >>> Vector(1, 2, 3).replace(z=4)
            Vector(x=1, y=2, z=4)
        """
        # XXX What if I want to use None?
        return Vector(
            self.x if x is None else x,
            self.y if y is None else y,
            self.z if z is None else z)

    def floor(self):
        """
        Return the vector with the floor of each component. This is only useful
        for vectors containing floating point components::

            >>> Vector(0.5, -0.5, 1.9)
            Vector(0.0, -1.0, 1.0)
        """
        return Vector(
            int(math.floor(self.x)),
            int(math.floor(self.y)),
            int(math.floor(self.z)))

    def ceil(self):
        """
        Return the vector with the ceiling of each component. This is only
        useful for vectors containing floating point components::

            >>> Vector(0.5, -0.5, 1.2)
            Vector(1.0, 0.0, 2.0)
        """
        return Vector(
            int(math.ceil(self.x)),
            int(math.ceil(self.y)),
            int(math.ceil(self.z)))

    def round(self, ndigits=0):
        """
        Return the vector with the rounded value of each component. This is
        only useful for vectors containing floating point components::

            >>> Vector(0.5, -0.5, 1.2)
            Vector(1.0, -1.0, 1.0)

        The *ndigits* argument operates as it does in the built-in
        :func:`round` function, specifying the number of decimal (or integer)
        places to round to.
        """
        return Vector(
            round(self.x, ndigits),
            round(self.y, ndigits),
            round(self.z, ndigits))

    def dot(self, other):
        """
        Return the `dot product`_ of the vector with the *other* vector. The
        result is a scalar value. For example::

            >>> Vector(1, 2, 3).dot(Vector(2, 2, 2))
            12
            >>> Vector(1, 2, 3).dot(X)
            1

        .. _dot product: http://en.wikipedia.org/wiki/Dot_product
        """
        return self.x * other.x + self.y * other.y + self.z * other.z

    def cross(self, other):
        """
        Return the `cross product`_ of the vector with the *other* vector. The
        result is another vector. For example::

            >>> Vector(1, 2, 3).cross(Vector(2, 2, 2))
            Vector(x=-2, y=4, z=-2)
            >>> Vector(1, 2, 3).cross(X)
            Vector(x=0, y=3, z=-2)

        .. _cross product: http://en.wikipedia.org/wiki/Cross_product
        """
        return Vector(
                self.y * other.z - self.z * other.y,
                self.z * other.x - self.x * other.z,
                self.x * other.y - self.y * other.x)

    def distance_to(self, other):
        """
        Return the Euclidian distance between two three dimensional points
        (represented as vectors), calculated according to `Pythagoras'
        theorem`_. For example::

            >>> Vector(1, 2, 3).distance_to(Vector(2, 2, 2))
            1.4142135623730951
            >>> Vector().distance_to(X)
            1.0

        .. _Pythagoras' theorem: http://en.wikipedia.org/wiki/Pythagorean_theorem
        """
        return (other - self).magnitude

    def angle_between(self, other):
        """
        Returns the angle between this vector and the *other* vector on a plane
        that contains both vectors. The result is measured in degrees. For
        example::

            >>> X.angle_between(Y)
            90.0
            >>> (X + Y).angle_between(X)
            45.00000000000001
        """
        return math.degrees(math.acos(self.unit.dot(other.unit)))

    def project(self, other):
        """
        Return the `scalar projection`_ of this vector onto the *other* vector.
        This is a scalar indicating the length of this vector in the direction
        of the *other* vector. For example::

            >>> Vector(1, 2, 3).project(2 * Y)
            2.0
            >>> Vector(3, 4, 5).project(Vector(3, 4, 0))
            5.0

        .. _scalar projection: https://en.wikipedia.org/wiki/Scalar_projection
        """
        return self.dot(other.unit)

    def rotate(self, angle, about, origin=None):
        """
        Return this vector after `rotation`_ of *angle* degrees about the line
        passing through *origin* in the direction *about*. Origin defaults to
        the vector 0, 0, 0. Hence, if this parameter is omitted this method
        calculates rotation about the axis (through the origin) defined by
        *about*.  For example::

            >>> Y.rotate(90, about=X)
            Vector(x=0, y=6.123233995736766e-17, z=1.0)
            >>> Vector(3, 4, 5).rotate(30, about=X, origin=10 * Y)
            Vector(x=3.0, y=2.3038475772933684, z=1.330127018922194)

        Information about rotation around arbitrary lines was obtained from
        `Glenn Murray's informative site`_.

        .. _rotation: https://en.wikipedia.org/wiki/Rotation_group_SO%283%29
        .. _Glenn Murray's informative site: http://inside.mines.edu/fs_home/gmurray/ArbitraryAxisRotation/
        """
        r = math.radians(angle)
        sin = math.sin(r)
        cos = math.cos(r)
        x, y, z = self
        if origin is None:
            # Fast-paths: rotation about a specific unit axis
            if about == X:
                return Vector(x, y * cos - z * sin, y * sin + z * cos)
            elif about == Y:
                return Vector(z * sin + x * cos, y, z * cos - x * sin)
            elif about == Z:
                return Vector(x * cos - y * sin, x * sin + y * cos, z)
            elif about == negX:
                return Vector(x, y * cos + z * sin, z * cos - y * sin)
            elif about == negY:
                return Vector(x * cos - z * sin, y, z * cos + x * sin)
            elif about == negZ:
                return Vector(x * cos + y * sin, y * cos - x * sin, z)
            # Rotation about an arbitrary axis
            u, v, w = about.unit
            return Vector(
                u * (u * x + v * y + w * z) * (1 - cos) + x * cos + (-w * y + v * z) * sin,
                v * (u * x + v * y + w * z) * (1 - cos) + y * cos + ( w * x - u * z) * sin,
                w * (u * x + v * y + w * z) * (1 - cos) + z * cos + (-v * x + u * y) * sin)
        # Rotation about an arbitrary line
        a, b, c = origin
        u, v, w = about.unit
        return Vector(
            (a * (v ** 2 + w ** 2) - u * (b * v + c * w - u * x - v * y - w * z)) * (1 - cos) + x * cos + (-c * v + b * w - w * y + v * z) * sin,
            (b * (u ** 2 + w ** 2) - v * (a * u + c * w - u * x - v * y - w * z)) * (1 - cos) + y * cos + ( c * u - a * w + w * x - u * z) * sin,
            (c * (u ** 2 + v ** 2) - w * (a * u + b * v - u * x - v * y - w * z)) * (1 - cos) + z * cos + (-b * u + a * v - v * x + u * y) * sin)

    @property
    def magnitude(self):
        """
        Returns the magnitude of the vector. This could also be considered the
        distance of the vector from the origin, i.e. ``v.magnitude`` is
        equivalent to ``Vector().distance_to(v)``. For example::

            >>> Vector(2, 4, 4).magnitude
            6.0
            >>> Vector().distance_to(Vector(2, 4, 4))
            6.0
        """
        return math.sqrt(self.x ** 2 + self.y ** 2 + self.z ** 2)

    @property
    def unit(self):
        """
        Return a `unit vector`_ (a vector with a magnitude of one) with the
        same direction as this vector::

            >>> X.unit
            Vector(x=1.0, y=0.0, z=0.0)
            >>> (2 * Y).unit
            Vector(x=0.0, y=1.0, z=0.0)

        .. note::

            If the vector's magnitude is zero, this property returns the
            original vector.

        .. _unit vector: http://en.wikipedia.org/wiki/Unit_vector
        """
        try:
            return self / self.magnitude
        except ZeroDivisionError:
            return self


# Short-hand variants
V = Vector
O = V()
X = V(x=1)
Y = V(y=1)
Z = V(z=1)
# These aren't exposed as short-hands; they're only pre-calculated here to
# speed up the fast-paths in the rotate() method
negX = V(x=-1)
negY = V(y=-1)
negZ = V(z=-1)


# TODO Yes, I'm being lazy with total_ordering ... probably ought to define all
# six comparison methods but I haven't got time right now ...

@total_ordering
class vector_range(Sequence):
    """
    Like :func:`range`, :class:`vector_range` is actually a type which
    efficiently represents a range of vectors. The arguments to the constructor
    must be :class:`Vector` instances (or objects which have integer ``x``,
    ``y``, and ``z`` attributes).

    If *step* is omitted, it defaults to ``Vector(1, 1, 1)``. If the *start*
    argument is omitted, it defaults to ``Vector(0, 0, 0)``. If any element
    of the *step* vector is zero, :exc:`ValueError` is raised.

    The contents of the range are largely determined by the *step* and *order*
    which specifies the order in which the axes of the range will be
    incremented.  For example, with the order ``'xyz'``, the X-axis will be
    incremented first, followed by the Y-axis, and finally the Z-axis. So, for
    a range with the default *start*, *step*, and *stop* set to ``Vector(3, 3,
    3)``, the contents of the range will be::

        >>> list(vector_range(Vector(3, 3, 3), order='xyz'))
        [Vector(0, 0, 0), Vector(1, 0, 0), Vector(2, 0, 0),
         Vector(0, 1, 0), Vector(1, 1, 0), Vector(2, 1, 0),
         Vector(0, 2, 0), Vector(1, 2, 0), Vector(2, 2, 0),
         Vector(0, 0, 1), Vector(1, 0, 1), Vector(2, 0, 1),
         Vector(0, 1, 1), Vector(1, 1, 1), Vector(2, 1, 1),
         Vector(0, 2, 1), Vector(1, 2, 1), Vector(2, 2, 1),
         Vector(0, 0, 2), Vector(1, 0, 2), Vector(2, 0, 2),
         Vector(0, 1, 2), Vector(1, 1, 2), Vector(2, 1, 2),
         Vector(0, 2, 2), Vector(1, 2, 2), Vector(2, 2, 2)]

    Vector ranges implement all common sequence operations except concatenation
    and repetition (due to the fact that range objects can only represent
    sequences that follow a strict pattern and repetition and concatenation
    usually cause the resulting sequence to violate that pattern).

    Vector ranges are extremely efficient compared to an equivalent
    :func:`list` or :func:`tuple` as they take a small (fixed) amount of
    memory, storing only the arguments passed in its construction and
    calculating individual items and sub-ranges as requested.

    Vector range objects implement the :class:`collections.Sequence` ABC,
    and provide features such as containment tests, element index lookup,
    slicing and support for negative indices.

    The default order (``'zxy'``) may seem an odd choice. This is primarily
    used as it's the order used by the Raspberry Juice server when returning
    results from the :ref:`world.getBlocks` call. In turn, Raspberry Juice
    probably uses this order as it results in returning a horizontal layer of
    vectors at a time (given the Y-axis is used for height in the Minecraft
    world).

    .. warning::

        Bear in mind that the ordering of a vector range may have a bearing on
        tests for its ordering and equality. Two ranges with different orders
        are unlikely to test equal even though they may have the same *start*,
        *stop*, and *step* attributes (and thus contain the same vectors, but
        in a different order).

    Vector ranges can be accessed by integer index, by :class:`Vector` index,
    or by a slice of vectors. For example::

        >>> v = vector_range(Vector() + 1, Vector() + 3)
        >>> list(v)
        [Vector(x=1, y=1, z=1),
         Vector(x=1, y=1, z=2),
         Vector(x=2, y=1, z=1),
         Vector(x=2, y=1, z=2),
         Vector(x=1, y=2, z=1),
         Vector(x=1, y=2, z=2),
         Vector(x=2, y=2, z=1),
         Vector(x=2, y=2, z=2)]
        >>> v[0]
        Vector(x=1, y=1, z=1)
        >>> v[Vector(0, 0, 0)]
        Vector(x=1, y=1, z=1)
        >>> v[Vector(1, 0, 0)]
        Vector(x=2, y=1, z=1)
        >>> v[-1]
        Vector(x=2, y=2, z=2)
        >>> v[Vector() - 1]
        Vector(x=2, y=2, z=2)
        >>> v[Vector(x=1):]
        vector_range(Vector(x=2, y=1, z=1), Vector(x=3, y=3, z=3),
                Vector(x=1, y=1, z=1), order='zxy')
        >>> list(v[Vector(x=1):])
        [Vector(x=2, y=1, z=1),
         Vector(x=2, y=1, z=2),
         Vector(x=2, y=2, z=1),
         Vector(x=2, y=2, z=2)]

    However, integer slices are not currently permitted.
    """

    def __init__(
            self, start, stop=None, step=None, order='zxy'):
        if stop is None:
            start, stop = Vector(), start
        if step is None:
            step = Vector(1, 1, 1)
        if (start != start // 1) or (stop != stop // 1) or (step != step // 1):
            raise TypeError('integer vectors are required')
        if order not in ('xyz', 'xzy', 'yxz', 'yzx', 'zxy', 'zyx'):
            raise ValueError('invalid order: %s' % order)
        if not (step.x and step.y and step.z):
            raise ValueError('no element of step may be zero')
        self._start = start
        self._stop = stop
        self._step = step
        self._order = order
        self._ranges = [
            range(
                getattr(self.start, axis),
                getattr(self.stop, axis),
                getattr(self.step, axis))
            for axis in order
            ]
        self._indexes = [
            order.index(axis)
            for axis in 'xyz'
            ]
        self._xrange, self._yrange, self._zrange = (
            self._ranges[i] for i in self._indexes
            )
        self._len = len(self._xrange) * len(self._yrange) * len(self._zrange)

    @property
    def start(self):
        return self._start

    @property
    def stop(self):
        return self._stop

    @property
    def step(self):
        return self._step

    @property
    def order(self):
        return self._order

    def __repr__(self):
        if self.start == Vector() and self.step == Vector() + 1:
            return 'vector_range(%r, order=%r)' % (self.stop, self.order)
        elif self.step == Vector() + 1:
            return 'vector_range(%r, %r, order=%r)' % (
                    self.start, self.stop, self.order)
        else:
            return 'vector_range(%r, %r, %r, order=%r)' % (
                    self.start, self.stop, self.step, self.order)

    def __len__(self):
        return self._len

    def __lt__(self, other):
        for v1, v2 in zip_longest(self, other):
            if v1 < v2:
                return True
            elif v1 > v2:
                return False
        return False

    def __eq__(self, other):
        # Fast-path: if the other object is an identical vector_range we
        # can quickly test whether we're equal
        if isinstance(other, vector_range):
            return (
                    self._xrange == other._xrange and
                    self._yrange == other._yrange and
                    self._zrange == other._zrange and
                    self.order == other.order
                    )
        # Normal case: test every element in each sequence
        for v1, v2 in zip_longest(self, other):
            if v1 != v2:
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def __reversed__(self):
        for i in reversed(range(len(self))):
            yield self[i]

    def __contains__(self, value):
        try:
            self.index(value)
        except ValueError:
            return False
        else:
            return True

    def __bool__(self):
        return len(self) > 0

    # Py2 compat
    __nonzero__ = __bool__

    def __getitem__(self, index):
        if isinstance(index, slice):
            return self._get_slice(index)
        elif isinstance(index, Vector):
            try:
                return Vector(*(
                    self._ranges[i][j]
                    for i, j in zip(self._indexes, index)
                    ))
            except IndexError:
                raise IndexError('list index out of range')
        else:
            if index < 0:
                index += len(self)
            if not (0 <= index < len(self)):
                raise IndexError('list index out of range')
            v = (
                self._ranges[0][index % len(self._ranges[0])],
                self._ranges[1][(index // len(self._ranges[0])) % len(self._ranges[1])],
                self._ranges[2][index // (len(self._ranges[0]) * len(self._ranges[1]))],
                )
            return Vector(*(v[i] for i in self._indexes))

    def _get_slice(self, s):
        try:
            step = Vector() + 1 if s.step is None else s.step
            start = Vector(None, None, None) if s.start is None else s.start
            stop = Vector(None, None, None) if s.stop is None else s.stop
            if not (step.x and step.y and step.z):
                raise ValueError(
                    "every element of the slice's step must be non-zero")
            x_range = self._xrange[slice(start.x, stop.x, step.x)]
            y_range = self._yrange[slice(start.y, stop.y, step.y)]
            z_range = self._zrange[slice(start.z, stop.z, step.z)]
        except AttributeError:
            raise ValueError(
                "vector_range slices must be composed of Vectors")
        return vector_range(
            Vector(x_range.start, y_range.start, z_range.start),
            Vector(x_range.stop, y_range.stop, z_range.stop),
            Vector(x_range.step, y_range.step, z_range.step),
            self.order)

    def index(self, value):
        """
        Return the zero-based index of *value* within the range, or raise
        :exc:`ValueError` if *value* does not exist in the range.
        """
        ranges = self._ranges
        i, j, k = (getattr(value, axis) for axis in self.order)
        try:
            i_indexes = set(rmod(len(ranges[0]), ranges[0].index(i), range(len(self))))
            j_indexes = set(
                    b
                    for a in rmod(len(ranges[1]), ranges[1].index(j),
                        range(len(self) // len(ranges[0])))
                    for b in rdiv(len(ranges[0]), a)
                    )
            k_indexes = set(rdiv(len(ranges[0]) * len(ranges[1]), ranges[2].index(k)))
            result = i_indexes & j_indexes & k_indexes
            assert len(result) == 1
            result = next(iter(result))
        except ValueError:
            raise ValueError('%r is not in range' % (value,))
        else:
            return result

    def count(self, value):
        """
        Return the count of instances of *value* within the range (note this
        can only be 0 or 1 in the case of a range, and thus is equivalent to
        testing membership with ``in``).
        """
        # count is provided by the ABC but inefficiently; given no vectors in
        # the range can be duplicated we provide a more efficient version here
        if value in self:
            return 1
        else:
            return 0


def rmod(denom, result, num_range):
    """
    Calculates the inverse of a mod operation.

    The *denom* parameter specifies the denominator of the original mod (%)
    operation. In this implementation, *denom* must be greater than 0. The
    *result* parameter specifies the result of the mod operation. For obvious
    reasons this value must be in the range ``[0, denom)`` (greater than or
    equal to zero and less than the denominator).

    Finally, *num_range* specifies the range that the numerator of the original
    mode operation can lie in. This must be an iterable sorted in ascending
    order with unique values (e.g. most typically a :func:`range`).

    The function returns the set of potential numerators (guaranteed to be a
    subset of *num_range*).
    """
    if denom <= 0:
        raise ValueError('invalid denominator')
    if not (0 <= result < denom):
        return set()
    if len(num_range) == 0:
        return set()
    assert num_range[-1] >= num_range[0]
    start = num_range[0] + (result - num_range[0] % denom) % denom
    stop = num_range.stop
    return range(start, stop, denom)


def rdiv(denom, result):
    """
    Calculates the inverse of a div operation.

    The *denom* parameter specifies the denominator of the original div (//)
    operation. In this implementation, *denom* must be greater than 0. The
    *result* parameter specifies the result of the div operation.

    The function returns the set of potential numerators.
    """
    if denom <= 0:
        raise ValueError('invalid denominator')
    return range(result * denom, result * denom + denom)


def sign(v):
    """
    Returns the sign of v as -1, 0, or 1; works for scalar values or
    :class:`Vector` instances.
    """
    try:
        return Vector(sign(v.x), sign(v.y), sign(v.z))
    except AttributeError:
        return 1 if v > 0 else -1 if v < 0 else 0


def line(start, end):
    """
    A three-dimensional implementation of `Bresenham's line algorithm`_,
    derived largely from `Bob Pendelton's implementation`_ (public domain).
    Given the end points of the line as the *start* and *end* vectors, this
    generator function yields the coordinate of each block (inclusive of the
    *start* and *end* vectors) that should be filled in to render the line.

    For example::

        >>> list(line(O, V(10, 5, 0)))
        [Vector(x=0, y=0, z=0),
         Vector(x=1, y=1, z=0),
         Vector(x=2, y=1, z=0),
         Vector(x=3, y=2, z=0),
         Vector(x=4, y=2, z=0),
         Vector(x=5, y=3, z=0),
         Vector(x=6, y=3, z=0),
         Vector(x=7, y=4, z=0),
         Vector(x=8, y=4, z=0),
         Vector(x=9, y=5, z=0),
         Vector(x=10, y=5, z=0)]

    .. _Bresenham's line algorithm: https://en.wikipedia.org/wiki/Bresenham%27s_line_algorithm
    .. _Bob Pendelton's implementation: ftp://ftp.isc.org/pub/usenet/comp.sources.unix/volume26/line3d
    """
    delta = end - start
    # Calculate the amount to increment each axis by; only the dominant axis
    # will advance by this amount on *every* iteration. Other axes will only
    # increment when the error demands it
    pos_inc = sign(delta)
    # Set up a vector containing the error incrementor. This will be added to
    # values tracking the axis error on each iteration
    error_inc = abs(delta) << 1
    # Calculate the subordinate and dominant axes. The dominant axis is simply
    # the one in which we must move furthest
    sub_axis1, sub_axis2, dominant_axis = sorted(
            'xyz', key=lambda axis: getattr(error_inc, axis))
    # Set up the error decrementor. This will be subtracted from the error
    # values when they turn positive (indicating that the corresponding axis
    # should advance)
    error_dec = getattr(error_inc, dominant_axis)
    # Set up a vector to track the error (this is only really required for the
    # subordinate axes)
    error = error_inc - (error_dec >> 1)
    # Convert vectors to dicts for the remainder of the algorithm
    error = error._asdict()
    error_inc = error_inc._asdict()
    pos_inc = pos_inc._asdict()
    pos = start._asdict()
    end = getattr(end, dominant_axis)
    while True:
        yield Vector(**pos)
        if pos[dominant_axis] == end:
            break
        pos[dominant_axis] += pos_inc[dominant_axis]
        if error[sub_axis1] >= 0:
            pos[sub_axis1] += pos_inc[sub_axis1]
            error[sub_axis1] -= error_dec
        error[sub_axis1] += error_inc[sub_axis1]
        if error[sub_axis2] >= 0:
            pos[sub_axis2] += pos_inc[sub_axis2]
            error[sub_axis2] -= error_dec
        error[sub_axis2] += error_inc[sub_axis2]


def lines(points, closed=True):
    """
    Extension of the :func:`line` function which returns all vectors necessary
    to render the lines connecting the specified *points* (which is an iterable
    of :class:`Vector` instances).

    If the optional *closed* parameter is ``True`` (the default) the last point
    in the *points* sequence will be connected to the first point. Otherwise,
    the lines will be left disconnected (assuming the last point is not
    coincident with the first). For example::

        >>> points = [O, 4*X, 4*Z]
        >>> list(lines(points))
        [Vector(x=0, y=0, z=0),
         Vector(x=1, y=0, z=0),
         Vector(x=2, y=0, z=0),
         Vector(x=3, y=0, z=0),
         Vector(x=4, y=0, z=0),
         Vector(x=3, y=0, z=1),
         Vector(x=2, y=0, z=2),
         Vector(x=1, y=0, z=3),
         Vector(x=0, y=0, z=4),
         Vector(x=0, y=0, z=3),
         Vector(x=0, y=0, z=2),
         Vector(x=0, y=0, z=1),
         Vector(x=0, y=0, z=0)]
    """
    first = None
    start = None
    for point in points:
        if start is None:
            first = point
            yield first
        else:
            for v in islice(line(start, point), 1, None):
                yield v
        start = point
    if closed and first != point:
        for v in islice(line(point, first), 1, None):
            yield v

def circle(center, radius, plane=Y):
    """
    Generates the coordinates of a three-dimensional circle centered at the
    vector *center*. The *radius* parameter is a vector specifying the distance
    of the circumference from the center. The optional *plane* parameter
    specifies another vector which, in combination with the *radius* vector,
    gives the plane that the circle exists within.

    For example, to create a circle centered at (0, 10, 0), with a radius of 5
    units, existing in the X-Y plane::

        >>> circle(10*Y, 5*X, Y)

    To create another circle with the same center and radius, but existing in
    the X-Z (ground) plane::

        >>> circle(10*Y, 5*X, Z)

    The algorithm used by this function is based on the `differences of roots`_
    method, extended to three dimensions.
    """
    if radius.angle_between(plane) != 90:
        plane = radius.cross(-(radius.cross(plane)))
    perp = plane.unit
    r = radius.magnitude**2
    last_points = None
    result = set()
    for radial_point in line(-radius, radius):
        circum_v = (perp * math.sqrt(r - radial_point.magnitude**2)).floor()
        top_point = (radial_point + circum_v)
        bottom_point = (radial_point - circum_v)
        if last_points is not None:
            for p in line(last_points[0], top_point):
                result.add(center + p)
            for p in line(last_points[1], bottom_point):
                result.add(center + p)
        last_points = (top_point, bottom_point)
    return result




# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# An alternate Python Minecraft library for the Rasperry-Pi
# Copyright (c) 2013-2015 Dave Jones <dave@waveform.org.uk>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
str = type('')


import pytest
import io
import picraft.player
from picraft import HostPlayer, Player, Vector, NoResponse, NotSupported
try:
    from unittest import mock
except ImportError:
    import mock


def test_players_len():
    conn = mock.MagicMock()
    conn.transact.return_value = '1|2|3'
    players = picraft.player.Players(conn)
    assert len(players) == 3

def test_players_contains():
    conn = mock.MagicMock()
    conn.transact.return_value = '1|2|3'
    players = picraft.player.Players(conn)
    assert 1 in players
    assert 4 not in players

def test_players_iter():
    conn = mock.MagicMock()
    conn.transact.return_value = '1|2|3'
    players = picraft.player.Players(conn)
    assert len(list(iter(players))) == 3

def test_players_get():
    conn = mock.MagicMock()
    def mock_transact(s):
        if s == 'world.getPlayerIds()':
            return '1|2|3'
        elif s == 'world.getPlayerId(foo)':
            return '1'
        else:
            raise NoResponse()
    conn.transact.side_effect = mock_transact
    conn.server_version = 'raspberry-juice'
    players = picraft.player.Players(conn)
    assert players[2].player_id == 2
    with pytest.raises(KeyError):
        players[4]
    assert players['foo'].player_id == 1
    with pytest.raises(KeyError):
        players['bar']

def test_players_keys():
    conn = mock.MagicMock()
    conn.transact.return_value = '1|2|3'
    players = picraft.player.Players(conn)
    assert set(players.keys()) == {1, 2, 3}

def test_players_values():
    conn = mock.MagicMock()
    conn.transact.return_value = '1|2|3'
    players = picraft.player.Players(conn)
    assert {p.player_id for p in players.values()} == {1, 2, 3}

def test_players_items():
    conn = mock.MagicMock()
    conn.transact.return_value = '1|2|3'
    players = picraft.player.Players(conn)
    assert {(k, v.player_id) for (k, v) in players.items()} == {(1, 1), (2, 2), (3, 3)}

def test_player_pos():
    conn = mock.MagicMock()
    conn.transact.return_value = '0.0,0.0,0.0'
    assert Player(conn, 1).pos == Vector(0.0, 0.0, 0.0)
    conn.transact.assert_called_once_with('entity.getPos(1)')
    Player(conn, 1).pos = Vector(1, 2, 3)
    conn.send.assert_called_once_with('entity.setPos(1,1,2,3)')

def test_player_tile_pos():
    conn = mock.MagicMock()
    conn.transact.return_value = '0,0,0'
    assert Player(conn, 1).tile_pos == Vector(0, 0, 0)
    conn.transact.assert_called_once_with('entity.getTile(1)')
    Player(conn, 1).tile_pos = Vector(1, 2, 3)
    conn.send.assert_called_once_with('entity.setTile(1,1,2,3)')

def test_player_heading():
    conn = mock.MagicMock()
    conn.server_version = 'raspberry-juice'
    conn.transact.return_value = '90.0'
    assert Player(conn, 1).heading == 90.0
    conn.transact.assert_called_once_with('entity.getRotation(1)')

def test_player_pitch():
    conn = mock.MagicMock()
    conn.server_version = 'raspberry-juice'
    conn.transact.return_value = '10.0'
    assert Player(conn, 1).pitch == 10.0
    conn.transact.assert_called_once_with('entity.getPitch(1)')

def test_player_direction():
    conn = mock.MagicMock()
    conn.server_version = 'raspberry-juice'
    conn.transact.return_value = '1.0,0.0,0.0'
    assert Player(conn, 1).direction == Vector(1.0, 0.0, 0.0)
    conn.transact.assert_called_once_with('entity.getDirection(1)')

def test_player_mcpi():
    conn = mock.MagicMock()
    conn.server_version = 'minecraft-pi'
    with pytest.raises(NotSupported):
        Player(conn, 1).heading
    with pytest.raises(NotSupported):
        Player(conn, 1).pitch
    with pytest.raises(NotSupported):
        Player(conn, 1).direction

def test_host_player_autojump():
    conn = mock.MagicMock()
    conn.server_version = 'minecraft-pi'
    with pytest.raises(AttributeError):
        HostPlayer(conn).autojump
    HostPlayer(conn).autojump = True
    conn.send.assert_called_once_with('player.setting(autojump,1)')
    conn.server_version = 'raspberry-juice'
    with pytest.raises(NotSupported):
        HostPlayer(conn).autojump = True

def test_host_player_pos():
    conn = mock.MagicMock()
    conn.transact.return_value = '0.0,0.0,0.0'
    assert HostPlayer(conn).pos == Vector(0.0, 0.0, 0.0)
    conn.transact.assert_called_once_with('player.getPos()')
    HostPlayer(conn).pos = Vector(1, 2, 3)
    conn.send.assert_called_once_with('player.setPos(1,2,3)')

def test_host_player_tile_pos():
    conn = mock.MagicMock()
    conn.transact.return_value = '0,0,0'
    assert HostPlayer(conn).tile_pos == Vector(0, 0, 0)
    conn.transact.assert_called_once_with('player.getTile()')
    HostPlayer(conn).tile_pos = Vector(1, 2, 3)
    conn.send.assert_called_once_with('player.setTile(1,2,3)')

def test_host_player_heading():
    conn = mock.MagicMock()
    conn.server_version = 'raspberry-juice'
    conn.transact.return_value = '90.0'
    assert HostPlayer(conn).heading == 90.0
    conn.transact.assert_called_once_with('player.getRotation()')

def test_host_player_pitch():
    conn = mock.MagicMock()
    conn.server_version = 'raspberry-juice'
    conn.transact.return_value = '10.0'
    assert HostPlayer(conn).pitch == 10.0
    conn.transact.assert_called_once_with('player.getPitch()')

def test_host_player_direction():
    conn = mock.MagicMock()
    conn.server_version = 'raspberry-juice'
    conn.transact.return_value = '1.0,0.0,0.0'
    assert HostPlayer(conn).direction == Vector(1.0, 0.0, 0.0)
    conn.transact.assert_called_once_with('player.getDirection()')




# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# An alternate Python Minecraft library for the Rasperry-Pi
# Copyright (c) 2013-2015 Dave Jones <dave@waveform.org.uk>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
str = type('')


import pytest
import warnings
import io
import picraft.block
from picraft import Block, Vector, vector_range, EmptySliceWarning
try:
    from unittest import mock
except ImportError:
    import mock


def test_read_block_data_filename():
    with mock.patch('io.open'):
        io.open.return_value = [b'1 0 1 1 test Test']
        assert list(picraft.block._read_block_data('foo.txt')) == [(1, 0, True, True, 'test', 'Test')]

def test_read_block_data_file_obj():
    data = [b'1 0 1 1 test Test']
    assert list(picraft.block._read_block_data(data)) == [(1, 0, True, True, 'test', 'Test')]

def test_read_block_color_filename():
    with mock.patch('io.open'):
        io.open.return_value = [b'1 0 ffffff', b'1 1 000000']
        assert list(picraft.block._read_block_color('foo.txt')) == [(1, 0, (255, 255, 255)), (1, 1, (0, 0, 0))]

def test_block_init():
    assert Block(1, 1) == Block.from_id(1, 1)
    assert Block(b'stone') == Block.from_name('stone')
    assert Block('stone') == Block.from_name('stone')
    assert Block('air', 1) == Block.from_name('air', 1)
    assert Block('#ffffff') == Block.from_color('#ffffff')
    assert Block((0, 0, 0)) == Block.from_color((0, 0, 0))
    assert Block(id=1) == Block.from_id(1)
    assert Block(name='grass') == Block.from_name('grass')
    assert Block(color='#ffffff', exact=False) == Block.from_color('#ffffff')
    with pytest.raises(TypeError):
        Block()
    with pytest.raises(TypeError):
        Block(1, 2, 3)

def test_block_from_string():
    assert Block.from_string('1,1') == Block(1, 1)
    with pytest.raises(ValueError):
        Block.from_string('foo')
    with pytest.raises(ValueError):
        Block.from_string('1.0,2.0')

def test_block_from_name():
    assert Block.from_name(b'air') == Block(0, 0)
    assert Block.from_name('air') == Block(0, 0)
    assert Block.from_name('stone') == Block(1, 0)
    with pytest.raises(ValueError):
        Block.from_name('foobarbaz')

def test_block_from_color():
    assert Block.from_color(b'#ffffff') == Block(35, 0)
    assert Block.from_color('#ffffff') == Block(35, 0)
    assert Block.from_color((255, 255, 255)) == Block(35, 0)
    assert Block.from_color((0.0, 0.0, 0.0)) == Block(35, 15)
    with pytest.raises(ValueError):
        Block.from_color('white')
    with pytest.raises(ValueError):
        Block.from_color((1, 2))

def test_block_from_color_exact():
    with pytest.raises(ValueError):
        Block.from_color(b'#ffffff', exact=True)

def test_block_platforms():
    assert Block.from_name('air').pi
    assert Block.from_name('air').pocket
    assert not Block.from_name('piston').pi
    assert not Block.from_name('piston').pocket

def test_block_name():
    assert Block(0, 0).name == 'air'

def test_block_desc():
    assert Block(35, 15).description == 'Black Wool'
    assert Block(35, 30).description == 'White Wool'

def test_blocks_get_one():
    conn = mock.MagicMock()
    conn.transact.return_value = '1,1'
    assert picraft.block.Blocks(conn)[Vector(1, 2, 3)] == Block(1, 1)
    conn.transact.assert_called_once_with('world.getBlockWithData(1,2,3)')

def test_blocks_get_many():
    v_from = Vector(1, 2, 3)
    v_to = Vector(2, 3, 5)
    conn = mock.MagicMock()
    conn.transact.return_value = '1,1'
    assert picraft.block.Blocks(conn)[v_from:v_to] == [
            Block(1, 1) for v in vector_range(v_from, v_to)]
    for v in vector_range(v_from, v_to):
        conn.transact.assert_any_call(
                'world.getBlockWithData(%d,%d,%d)' % (v.x, v.y, v.z))

def test_blocks_get_many_fast():
    v_from = Vector(1, 2, 3)
    v_to = Vector(2, 3, 5)
    conn = mock.MagicMock()
    conn.server_version = 'raspberry-juice'
    conn.transact.return_value = '1,1'
    picraft.block.Blocks(conn)[v_from:v_to] == [
            Block(1, 0) for v in vector_range(v_from, v_to)]
    conn.transact.assert_called_once_with(
            'world.getBlocks(%s,%s)' % (v_from, v_to - 1))

def test_blocks_get_none():
    conn = mock.MagicMock()
    v_from = Vector(1, 2, 3)
    v_to = Vector(2, 3, 5)
    with warnings.catch_warnings():
        warnings.simplefilter('error')
        with pytest.raises(EmptySliceWarning):
            picraft.block.Blocks(conn)[v_to:v_from]

def test_blocks_set_none():
    conn = mock.MagicMock()
    with warnings.catch_warnings():
        warnings.simplefilter('error')
        with pytest.raises(EmptySliceWarning):
            picraft.block.Blocks(conn)[Vector(1, 1, 1):Vector(3, 3, -1)] = Block(0, 0)

def test_blocks_set_one():
    conn = mock.MagicMock()
    picraft.block.Blocks(conn)[Vector(1, 2, 3)] = Block(0, 0)
    conn.send.assert_called_once_with('world.setBlock(1,2,3,0,0)')

def test_blocks_set_many_same():
    conn = mock.MagicMock()
    v_from = Vector(1, 2, 3)
    v_to = Vector(2, 3, 5)
    picraft.block.Blocks(conn)[v_from:v_to] = Block(0, 0)
    conn.send.assert_called_once_with('world.setBlocks(1,2,3,1,2,4,0,0)')

def test_blocks_set_many_different():
    conn = mock.MagicMock()
    v_from = Vector(1, 2, 3)
    v_to = Vector(2, 3, 5)
    blocks = [Block(1, 1) for v in vector_range(v_from, v_to)]
    picraft.block.Blocks(conn)[v_from:v_to] = blocks
    for v in vector_range(v_from, v_to):
        conn.send.assert_any_call(
                'world.setBlock(%d,%d,%d,1,1)' % (v.x, v.y, v.z))
    with pytest.raises(ValueError):
        picraft.block.Blocks(conn)[v_from:v_to] = blocks[1:]
    with pytest.raises(ValueError):
        picraft.block.Blocks(conn)[v_from:v_to] = blocks * 2



# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# An alternate Python Minecraft library for the Rasperry-Pi
# Copyright (c) 2013-2015 Dave Jones <dave@waveform.org.uk>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
The exc module defines the various exception classes specific to picraft.

.. note::

    All items in this module are available from the :mod:`picraft` namespace
    without having to import :mod:`picraft.exc` directly.

The following items are defined in the module:


Exceptions
==========

.. autoexception:: Error

.. autoexception:: NotSupported

.. autoexception:: ConnectionError

.. autoexception:: ConnectionClosed

.. autoexception:: CommandError

.. autoexception:: NoResponse

.. autoexception:: BatchStarted

.. autoexception:: BatchNotStarted

.. autoexception:: EmptySliceWarning

"""

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
str = type('')

import socket


class Error(Exception):
    "Base class for all PiCraft exceptions"

class NotSupported(Error, NotImplementedError):
    "Exception raised for unimplemented methods / properties"

class ConnectionError(Error, socket.error):
    "Base class for PiCraft errors relating to network communications"

class ConnectionClosed(ConnectionError):
    "Exception raised when an operation is attempted against a closed connection"

class CommandError(ConnectionError):
    "Exception raised when a network command fails"

class NoResponse(ConnectionError):
    "Exception raised when a network command expects a response but gets none"

class BatchStarted(ConnectionError):
    "Exception raised when a batch is started before a prior one is complete"

class BatchNotStarted(ConnectionError):
    "Exception raised when a batch is terminated when none has been started"

class EmptySliceWarning(Warning):
    "Warning raised when a zero-length vector slice is passed to blocks"




# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# An alternate Python Minecraft library for the Rasperry-Pi
# Copyright (c) 2013-2015 Dave Jones <dave@waveform.org.uk>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
The block module defines the :class:`Block` class, which is used to represent
the type of a block and any associated data it may have, and the
:class:`Blocks` class which is used to implement the
:attr:`~picraft.world.World.blocks` attribute on the
:class:`~picraft.world.World` class.

.. note::

    All items in this module, except the compatibility constants, are available
    from the :mod:`picraft` namespace without having to import
    :mod:`picraft.block` directly.

The following items are defined in the module:

Block
=====

.. autoclass:: Block(id, data)


BLOCK_COLORS
============

.. data:: BLOCK_COLORS

    A set of the available colors that can be used with
    :meth:`Block.from_color`.  Each color is represented as ``(red, green,
    blue)`` tuple where each component is an integer between 0 and 255.


Compatibility
=============

Finally, the module also contains compatibility values equivalent to those
in the mcpi.block module of the reference implementation. Each value represents
the type of a block with no associated data:

===================  ====================  =====================
AIR                  FURNACE_ACTIVE        MUSHROOM_RED
BED                  FURNACE_INACTIVE      NETHER_REACTOR_CORE
BEDROCK              GLASS                 OBSIDIAN
BEDROCK_INVISIBLE    GLASS_PANE            REDSTONE_ORE
BOOKSHELF            GLOWING_OBSIDIAN      SAND
BRICK_BLOCK          GLOWSTONE_BLOCK       SANDSTONE
CACTUS               GOLD_BLOCK            SAPLING
CHEST                GOLD_ORE              SNOW
CLAY                 GRASS                 SNOW_BLOCK
COAL_ORE             GRASS_TALL            STAIRS_COBBLESTONE
COBBLESTONE          GRAVEL                STAIRS_WOOD
COBWEB               ICE                   STONE
CRAFTING_TABLE       IRON_BLOCK            STONE_BRICK
DIAMOND_BLOCK        IRON_ORE              STONE_SLAB
DIAMOND_ORE          LADDER                STONE_SLAB_DOUBLE
DIRT                 LAPIS_LAZULI_BLOCK    SUGAR_CANE
DOOR_IRON            LAPIS_LAZULI_ORE      TNT
DOOR_WOOD            LAVA                  TORCH
FARMLAND             LAVA_FLOWING          WATER
FENCE                LAVA_STATIONARY       WATER_FLOWING
FENCE_GATE           LEAVES                WATER_STATIONARY
FIRE                 MELON                 WOOD
FLOWER_CYAN          MOSS_STONE            WOOD_PLANKS
FLOWER_YELLOW        MUSHROOM_BROWN        WOOL
===================  ====================  =====================

Use these compatibility constants by importing the block module explicitly.
For example::

    >>> from picraft import block
    >>> block.AIR
    <Block "air" id=0 data=0>
    >>> block.TNT
    <Block "tnt" id=46 data=0>
"""

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
try:
    from itertools import izip as zip
except ImportError:
    pass
str = type('')


import io
import warnings
from math import sqrt
from collections import namedtuple
try:
    # Py2 compat
    from itertools import izip_longest as zip_longest
except ImportError:
    from itertools import zip_longest

from pkg_resources import resource_stream
from .exc import EmptySliceWarning
from .vector import Vector, vector_range


def _read_block_data(filename_or_object):
    if isinstance(filename_or_object, str):
        stream = io.open(filename_or_object, 'rb')
    else:
        stream = filename_or_object
    for line in stream:
        line = line.decode('utf-8').strip()
        if line and not line.startswith('#'):
            id, data, pi, pocket, name, description = line.split(None, 5)
            yield int(id), int(data), bool(int(pi)), bool(int(pocket)), name, description

def _read_block_color(filename_or_object):
    if isinstance(filename_or_object, str):
        stream = io.open(filename_or_object, 'rb')
    else:
        stream = filename_or_object
    int2color = lambda n: ((n & 0xff0000) >> 16, (n & 0xff00) >> 8, n & 0xff)
    for line in stream:
        line = line.decode('utf-8').strip()
        if line and not line.startswith('#'):
            id, data, color = line.split(None, 2)
            yield int(id), int(data), int2color(int(color, 16))


_BLOCKS_DB = {
    (id, data): (pi, pocket, name, description)
    for (id, data, pi, pocket, name, description) in
        _read_block_data(resource_stream(__name__, 'block.data'))
    }

_BLOCKS_BY_ID = {
    id: (pi, pocket, name)
    for (id, data), (pi, pocket, name, description) in _BLOCKS_DB.items()
    if data == 0
    }

_BLOCKS_BY_NAME = {
    name: id
    for (id, data), (pi, pocket, name, description) in _BLOCKS_DB.items()
    if data == 0
    }

_BLOCKS_BY_COLOR = {
    color: (id, data)
    for (id, data, color) in
        _read_block_color(resource_stream(__name__, 'block.color'))
    }

BLOCK_COLORS = _BLOCKS_BY_COLOR.keys()


class Block(namedtuple('Block', ('id', 'data'))):
    """
    Represents a block within the Minecraft world.

    Blocks within the Minecraft world are represented by two values: an *id*
    which defines the type of the block (air, stone, grass, wool, etc.) and an
    optional *data* value (defaults to 0) which means different things for
    different block types (e.g.  for wool it defines the color of the wool).

    Blocks are represented by this library as a :func:`namedtuple` of the *id*
    and the *data*. Calculated properties are provided to look up the
    :attr:`name` and :attr:`description` of the block from a database derived
    from the Minecraft wiki, and classmethods are defined to construct a block
    definition from an :meth:`id <from_id>` or from alternate things like a
    :meth:`name <from_name>` or an :meth:`RGB color <from_color>`::

        >>> Block.from_id(0, 0)
        <Block "air" id=0 data=0>
        >>> Block.from_id(2, 0)
        <Block "grass" id=2 data=0>
        >>> Block.from_name('stone')
        <Block "stone" id=1 data=0>
        >>> Block.from_color('#ffffff')
        <Block "wool" id=35 data=0>

    The default constructor attempts to guess which classmethod to call based
    on the following rules (in the order given):

    1. If the constructor is passed a string beginning with '#' that is 7
       characters long, it calls :meth:`from_color`

    2. If the constructor is passed a tuple containing 3 values, it calls
       :meth:`from_color`

    3. If the constructor is passed a string (not matching the case above)
       it calls :meth:`from_name`

    4. Otherwise the constructor calls :meth:`from_id` with all given
       parameters

    This means that the examples above can be more easily written::

        >>> Block(0, 0)
        <Block "air" id=0 data=0>
        >>> Block(2, 0)
        <Block "grass" id=2 data=0>
        >>> Block('stone')
        <Block "stone" id=1 data=0>
        >>> Block('#ffffff')
        <Block "wool" id=35 data=0>

    Aliases are provided for compatibility with the official reference
    implementation (AIR, GRASS, STONE, etc)::

        >>> import picraft.block
        >>> picraft.block.WATER
        <Block "flowing_water" id=8 data=0>

    .. automethod:: from_color

    .. automethod:: from_id

    .. automethod:: from_name

    .. attribute:: id

        The "id" or type of the block. Each block type in Minecraft has a
        unique value. For example, air blocks have the id 0, stone, has id 1,
        and so forth. Generally it is clearer in code to refer to a block's
        :attr:`name` but it may be quicker to use the id.

    .. attribute:: data

        Certain types of blocks have variants which are dictated by the data
        value associated with them. For example, the color of a wool block
        is determined by the *data* attribute (e.g. white is 0, red is 14,
        and so on).

    .. autoattribute:: pi

    .. autoattribute:: pocket

    .. autoattribute:: name

    .. autoattribute:: description
    """

    def __new__(cls, *args, **kwargs):
        if len(args) >= 1:
            a = args[0]
            if isinstance(a, bytes):
                a = a.decode('utf-8')
            if isinstance(a, str) and len(a) == 7 and a.startswith('#'):
                return cls.from_color(*args, **kwargs)
            elif isinstance(a, tuple) and len(a) == 3:
                return cls.from_color(*args, **kwargs)
            elif isinstance(a, str):
                return cls.from_name(*args, **kwargs)
            else:
                return cls.from_id(*args, **kwargs)
        else:
            if 'id' in kwargs:
                return cls.from_id(**kwargs)
            elif 'name' in kwargs:
                return cls.from_name(**kwargs)
            elif 'color' in kwargs:
                return cls.from_color(**kwargs)
        raise TypeError('invalid combination of arguments for Block')

    @classmethod
    def from_string(cls, s):
        id_, data = s.split(',', 1)
        return cls.from_id(int(id_), int(data))

    @classmethod
    def from_id(cls, id, data=0):
        """
        Construct a :class:`Block` instance from an *id* integer. This may be
        used to construct blocks in the classic manner; by specifying a number
        representing the block's type, and optionally a data value. For
        example::

            >>> from picraft import *
            >>> Block.from_id(1)
            <Block "stone" id=1 data=0>
            >>> Block.from_id(2, 0)
            <Block "grass" id=2 data=0>

        The optional *data* parameter defaults to 0. Note that calling the
        default constructor with an integer parameter is equivalent to calling
        this method::

            >>> Block(1)
            <Block "stone" id=1" data=0>
        """
        return super(Block, cls).__new__(cls, id, data)

    @classmethod
    def from_name(cls, name, data=0):
        """
        Construct a :class:`Block` instance from a *name*, as returned by the
        :attr:`name` property. This may be used to construct blocks in a more
        "friendly" way within code. For example::

            >>> from picraft import *
            >>> Block.from_name('stone')
            <Block "stone" id=1 data=0>
            >>> Block.from_name('wool', data=2)
            <Block "wool" id=35 data=2>

        The optional *data* parameter can be used to specify the data component
        of the new :class:`Block` instance; it defaults to 0. Note that calling
        the default constructor with a string that doesn't start with "#" is
        equivalent to calling this method::

            >>> Block('stone')
            <Block "stone" id=1 data=0>
        """
        if isinstance(name, bytes):
            name = name.decode('utf-8')
        try:
            id_ = _BLOCKS_BY_NAME[name]
        except KeyError:
            raise ValueError('unknown name %s' % name)
        return cls(id_, data)

    @classmethod
    def from_color(cls, color, exact=False):
        """
        Construct a :class:`Block` instance from a *color* which can be
        represented as:

        * A tuple of ``(red, green, blue)`` integer byte values between 0 and
          255
        * A tuple of ``(red, green, blue)`` float values between 0.0 and 1.0
        * A string in the format '#rrggbb' where rr, gg, and bb are hexadecimal
          representations of byte values.

        If *exact* is ``False`` (the default), and an exact match for the
        requested color cannot be found, the nearest color (determined simply
        by Euclidian distance) is returned. If *exact* is ``True`` and an exact
        match cannot be found, a :exc:`ValueError` will be raised::

            >>> from picraft import *
            >>> Block.from_color('#ffffff')
            <Block "wool" id=35 data=0>
            >>> Block.from_color('#ffffff', exact=True)
            Traceback (most recent call last):
              File "<stdin>", line 1, in <module>
              File "picraft/block.py", line 351, in from_color
                if exact:
            ValueError: no blocks match color #ffffff
            >>> Block.from_color((1, 0, 0))
            <Block "wool" id=35 data=14>

        Note that calling the default constructor with any of the formats
        accepted by this method is equivalent to calling this method::

            >>> Block('#ffffff')
            <Block "wool" id=35 data=0>
        """
        if isinstance(color, bytes):
            color = color.decode('utf-8')
        if isinstance(color, str):
            try:
                if not (color.startswith('#') and len(color) == 7):
                    raise ValueError()
                color = (
                        int(color[1:3], 16),
                        int(color[3:5], 16),
                        int(color[5:7], 16))
            except ValueError:
                raise ValueError('unrecognized color format: %s' % color)
        else:
            try:
                r, g, b = color
            except (TypeError, ValueError):
                raise ValueError('expected three values in color')
            if 0.0 <= r <= 1.0 and 0.0 <= g <= 1.0 and 0.0 <= b <= 1.0:
                color = tuple(int(n * 255) for n in color)
        try:
            id_, data = _BLOCKS_BY_COLOR[color]
        except KeyError:
            r, g, b = color
            if exact:
                raise ValueError(
                    'no blocks match color #%06x' % (r << 16 | g << 8 | b))
            diff = lambda block_color: sqrt(
                    sum((c1 - c2) ** 2 for c1, c2 in zip(color, block_color)))
            matched_color = sorted(_BLOCKS_BY_COLOR, key=diff)[0]
            id_, data = _BLOCKS_BY_COLOR[matched_color]
        return cls(id_, data)

    def __repr__(self):
        try:
            return '<Block "%s" id=%d data=%d>' % (self.name, self.id, self.data)
        except KeyError:
            return '<Block id=%d data=%d>' % (self.id, self.data)

    @property
    def pi(self):
        """
        Returns a bool indicating whether the block is present in the Pi
        Edition of Minecraft.
        """
        return _BLOCKS_BY_ID[self.id][0]

    @property
    def pocket(self):
        """
        Returns a bool indicating whether the block is present in the Pocket
        Edition of Minecraft.
        """
        return _BLOCKS_BY_ID[self.id][1]

    @property
    def name(self):
        """
        Return the name of the block. This is a unique identifier string which
        can be used to construct a :class:`Block` instance with
        :meth:`from_name`.
        """
        return _BLOCKS_BY_ID[self.id][2]

    @property
    def description(self):
        """
        Return a description of the block. This string is not guaranteed to be
        unique and is only intended for human use.
        """
        try:
            return _BLOCKS_DB[(self.id, self.data)][3]
        except KeyError:
            return _BLOCKS_DB[(self.id, 0)][3]


class Blocks(object):
    """
    This class implements the :attr:`~picraft.world.World.blocks` attribute.
    """
    def __init__(self, connection):
        self._connection = connection

    def __repr__(self):
        return '<Blocks>'

    def __getitem__(self, index):
        if isinstance(index, slice):
            index = vector_range(index.start, index.stop, index.step)
        if isinstance(index, vector_range):
            vr = index
            if not vr:
                warnings.warn(EmptySliceWarning(
                    "ignoring empty slice passed to blocks"))
            elif (
                    abs(vr.step) == Vector(1, 1, 1) and
                    self._connection.server_version == 'raspberry-juice'):
                return [
                    Block.from_string('%d,0' % int(i))
                    for i in self._connection.transact(
                        'world.getBlocks(%d,%d,%d,%d,%d,%d)' % (
                        vr.start.x, vr.start.y, vr.start.z,
                        vr.stop.x - vr.step.x, vr.stop.y - vr.step.y, vr.stop.z - vr.step.z)).split(',')
                    ]
            else:
                return [
                    Block.from_string(
                        self._connection.transact(
                            'world.getBlockWithData(%d,%d,%d)' % (v.x, v.y, v.z)))
                    for v in vector_range(index.start, index.stop, index.step)
                    ]
        else:
            return Block.from_string(
                self._connection.transact(
                    'world.getBlockWithData(%d,%d,%d)' % (index.x, index.y, index.z)))

    def __setitem__(self, index, value):
        if isinstance(index, slice):
            index = vector_range(index.start, index.stop, index.step)
        if isinstance(index, vector_range):
            vr = index
            if not vr:
                warnings.warn(EmptySliceWarning(
                    "ignoring empty slice passed to blocks"))
            elif (
                    abs(vr.step) == Vector(1, 1, 1) and
                    hasattr(value, 'id') and
                    hasattr(value, 'data')):
                self._connection.send(
                    'world.setBlocks(%d,%d,%d,%d,%d,%d,%d,%d)' % (
                        vr.start.x, vr.start.y, vr.start.z,
                        vr.stop.x - vr.step.x, vr.stop.y - vr.step.y, vr.stop.z - vr.step.z,
                        value.id, value.data))
            else:
                for v, b in zip_longest(vr, value):
                    if v is None:
                        raise ValueError('too many blocks for vector range')
                    if b is None:
                        raise ValueError('not enough blocks for vector range')
                    self._connection.send(
                        'world.setBlock(%d,%d,%d,%d,%d)' % (
                            v.x, v.y, v.z, b.id, b.data))
        else:
            self._connection.send(
                'world.setBlock(%d,%d,%d,%d,%d)' % (
                    index.x, index.y, index.z, value.id, value.data))


AIR                 = Block(0)
STONE               = Block(1)
GRASS               = Block(2)
DIRT                = Block(3)
COBBLESTONE         = Block(4)
WOOD_PLANKS         = Block(5)
SAPLING             = Block(6)
BEDROCK             = Block(7)
WATER_FLOWING       = Block(8)
WATER               = WATER_FLOWING
WATER_STATIONARY    = Block(9)
LAVA_FLOWING        = Block(10)
LAVA                = LAVA_FLOWING
LAVA_STATIONARY     = Block(11)
SAND                = Block(12)
GRAVEL              = Block(13)
GOLD_ORE            = Block(14)
IRON_ORE            = Block(15)
COAL_ORE            = Block(16)
WOOD                = Block(17)
LEAVES              = Block(18)
GLASS               = Block(20)
LAPIS_LAZULI_ORE    = Block(21)
LAPIS_LAZULI_BLOCK  = Block(22)
SANDSTONE           = Block(24)
BED                 = Block(26)
COBWEB              = Block(30)
GRASS_TALL          = Block(31)
WOOL                = Block(35)
FLOWER_YELLOW       = Block(37)
FLOWER_CYAN         = Block(38)
MUSHROOM_BROWN      = Block(39)
MUSHROOM_RED        = Block(40)
GOLD_BLOCK          = Block(41)
IRON_BLOCK          = Block(42)
STONE_SLAB_DOUBLE   = Block(43)
STONE_SLAB          = Block(44)
BRICK_BLOCK         = Block(45)
TNT                 = Block(46)
BOOKSHELF           = Block(47)
MOSS_STONE          = Block(48)
OBSIDIAN            = Block(49)
TORCH               = Block(50)
FIRE                = Block(51)
STAIRS_WOOD         = Block(53)
CHEST               = Block(54)
DIAMOND_ORE         = Block(56)
DIAMOND_BLOCK       = Block(57)
CRAFTING_TABLE      = Block(58)
FARMLAND            = Block(60)
FURNACE_INACTIVE    = Block(61)
FURNACE_ACTIVE      = Block(62)
DOOR_WOOD           = Block(64)
LADDER              = Block(65)
STAIRS_COBBLESTONE  = Block(67)
DOOR_IRON           = Block(71)
REDSTONE_ORE        = Block(73)
SNOW                = Block(78)
ICE                 = Block(79)
SNOW_BLOCK          = Block(80)
CACTUS              = Block(81)
CLAY                = Block(82)
SUGAR_CANE          = Block(83)
FENCE               = Block(85)
GLOWSTONE_BLOCK     = Block(89)
BEDROCK_INVISIBLE   = Block(95)
STONE_BRICK         = Block(98)
GLASS_PANE          = Block(102)
MELON               = Block(103)
FENCE_GATE          = Block(107)
GLOWING_OBSIDIAN    = Block(246)
NETHER_REACTOR_CORE = Block(247)




#!/usr/bin/env python3
# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# An alternate Python Minecraft library for the Rasperry-Pi
# Copyright (c) 2013-2015 Dave Jones <dave@waveform.org.uk>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""An alternate Python interface for Minecraft."""

import os
import sys
from setuptools import setup, find_packages

if sys.version_info[0] == 2:
    if not sys.version_info >= (2, 7):
        raise ValueError('This package requires Python 2.7 or above')
elif sys.version_info[0] == 3:
    if not sys.version_info >= (3, 2):
        raise ValueError('This package requires Python 3.2 or above')
else:
    raise ValueError('What version of Python is this?!')

HERE = os.path.abspath(os.path.dirname(__file__))

# Workaround <http://www.eby-sarna.com/pipermail/peak/2010-May/003357.html>
try:
    import multiprocessing
except ImportError:
    pass

__project__      = 'picraft'
__version__      = '0.4'
__author__       = 'Dave Jones'
__author_email__ = 'dave@waveform.org.uk'
__url__          = 'http://picraft.readthedocs.org/'
__platforms__    = 'ALL'

__classifiers__ = [
    'Development Status :: 3 - Alpha',
    'Environment :: Console',
    'Environment :: X11 Applications',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: BSD License',
    'Operating System :: POSIX :: Linux',
    'Operating System :: Microsoft :: Windows',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3.2',
    'Programming Language :: Python :: 3.3',
    'Topic :: Games/Entertainment',
    ]

__keywords__ = [
    'raspberrypi',
    'minecraft',
    ]

__requires__ = [
    ]

__extra_requires__ = {
    'doc':   ['sphinx'],
    'test':  ['pytest', 'coverage', 'mock'],
    }

if sys.version_info[:2] == (3, 2):
    __extra_requires__['doc'].extend([
        # Particular versions are required for Python 3.2 compatibility. The
        # ordering is reversed because that's what easy_install needs...
        'Jinja2<2.7',
        'MarkupSafe<0.16',
        ])

__entry_points__ = {
    }


def main():
    import io
    with io.open(os.path.join(HERE, 'README.rst'), 'r') as readme:
        setup(
            name                 = __project__,
            version              = __version__,
            description          = __doc__,
            long_description     = readme.read(),
            classifiers          = __classifiers__,
            author               = __author__,
            author_email         = __author_email__,
            url                  = __url__,
            license              = [
                c.rsplit('::', 1)[1].strip()
                for c in __classifiers__
                if c.startswith('License ::')
                ][0],
            keywords             = __keywords__,
            packages             = find_packages(),
            include_package_data = True,
            platforms            = __platforms__,
            install_requires     = __requires__,
            extras_require       = __extra_requires__,
            entry_points         = __entry_points__,
            )

if __name__ == '__main__':
    main()

